"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import inspect
import warnings
import logging
import traceback
import sys
from typing import Union, Callable, List, Optional, Tuple, Any, Coroutine, Dict
from typing_extensions import Literal

from twitchio.errors import HTTPException
from . import models
from .websocket import WSConnection
from .http import TwitchHTTP
from .channel import Channel
from .message import Message
from .user import User, PartialUser, SearchUser
from .cache import user_cache, id_cache

__all__ = ("Client",)

logger = logging.getLogger("twitchio.client")


class Client:
    """TwitchIO Client object that is used to interact with the Twitch API and connect to Twitch IRC over websocket.

    Parameters
    ------------
    token: :class:`str`
        An OAuth Access Token to login with on IRC and interact with the API.
    client_secret: Optional[:class:`str`]
        An optional application Client Secret used to generate Access Tokens automatically.
    initial_channels: Optional[Union[:class:`list`, :class:`tuple`, Callable]]
        An optional list, tuple or callable which contains channel names to connect to on startup.
        If this is a callable, it must return a list or tuple.
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The event loop the client will use to run.
    heartbeat: Optional[float]
        An optional float in seconds to send a PING message to the server. Defaults to 30.0.
    retain_cache: Optional[bool]
        An optional bool that will retain the cache if PART is received from websocket when True.
        It will still remove from cache if part_channels is manually called. Defaults to True.

    Attributes
    ------------
    loop: :class:`asyncio.AbstractEventLoop`
        The event loop the Client uses.
    """

    def __init__(
        self,
        token: str,
        *,
        client_secret: str = None,
        initial_channels: Union[list, tuple, Callable] = None,
        loop: asyncio.AbstractEventLoop = None,
        heartbeat: Optional[float] = 30.0,
        retain_cache: Optional[bool] = True,
    ):
        self.loop: asyncio.AbstractEventLoop = loop or asyncio.get_event_loop()
        self._heartbeat = heartbeat

        token = token.replace("oauth:", "")

        self._http = TwitchHTTP(self, api_token=token, client_secret=client_secret)
        self._connection = WSConnection(
            client=self,
            token=token,
            loop=self.loop,
            initial_channels=initial_channels,
            heartbeat=heartbeat,
            retain_cache=retain_cache,
        )

        self._events = {}
        self._waiting: List[Tuple[str, Callable[[...], bool], asyncio.Future]] = []
        self.registered_callbacks: Dict[Callable, str] = {}
        self._closing: Optional[asyncio.Event] = None

    @classmethod
    def from_client_credentials(
        cls,
        client_id: str,
        client_secret: str,
        *,
        loop: asyncio.AbstractEventLoop = None,
        heartbeat: Optional[float] = 30.0,
    ) -> "Client":
        """
        creates a client application token from your client credentials.

        .. warning:

            this is not suitable for logging in to IRC.

        .. note:

            This classmethod skips :meth:`~.__init__`

        Parameters
        ------------
        client_id: :class:`str`

        client_secret: :class:`str`
            An application Client Secret used to generate Access Tokens automatically.
        loop: Optional[:class:`asyncio.AbstractEventLoop`]
            The event loop the client will use to run.

        Returns
        --------
        A new :class:`Client` instance
        """
        self = cls.__new__(cls)
        self.loop = loop or asyncio.get_event_loop()
        self._http = TwitchHTTP(self, client_id=client_id, client_secret=client_secret)
        self._heartbeat = heartbeat
        self._connection = WSConnection(
            client=self,
            loop=self.loop,
            initial_channels=None,
            heartbeat=self._heartbeat,
        )  # The only reason we're even creating this is to avoid attribute errors
        self._events = {}
        self._waiting = []
        self.registered_callbacks = {}
        return self

    def run(self):
        """
        A blocking function that starts the asyncio event loop,
        connects to the twitch IRC server, and cleans up when done.
        """
        try:
            task = self.loop.create_task(self.connect())
            self.loop.run_until_complete(task)  # this'll raise if the connect fails
            self.loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            if not self._closing.is_set():
                self.loop.run_until_complete(self.close())

            self.loop.close()

    async def start(self):
        """|coro|

        Connects to the twitch IRC server, and cleanly disconnects when done.
        """
        if self.loop is not asyncio.get_running_loop():
            raise RuntimeError(
                f"Attempted to start a {self.__class__.__name__} instance on a different loop "
                f"than the one it was initialized with."
            )
        try:
            await self.connect()
            await self._closing.wait()
        finally:
            if not self._closing.is_set():
                await self.close()

    async def connect(self):
        """|coro|

        Connects to the twitch IRC server
        """
        self._closing = asyncio.Event()
        await self._connection._connect()

    async def close(self):
        """|coro|

        Cleanly disconnects from the twitch IRC server
        """
        self._closing.set()
        await self._connection._close()

    def run_event(self, event_name, *args):
        name = f"event_{event_name}"
        logger.debug(f"dispatching event {event_name}")

        async def wrapped(func):
            try:
                await func(*args)
            except Exception as e:
                if name == "event_error":
                    # don't enter a dispatch loop!
                    raise

                self.run_event("error", e)

        inner_cb = getattr(self, name, None)
        if inner_cb is not None:
            if inspect.iscoroutinefunction(inner_cb):
                self.loop.create_task(wrapped(inner_cb))
            else:
                warnings.warn(
                    f"event '{name}' callback is not a coroutine",
                    category=RuntimeWarning,
                )

        if name in self._events:
            for event in self._events[name]:
                self.loop.create_task(wrapped(event))

        for e, check, future in self._waiting:
            if e == event_name:
                if check(*args):
                    future.set_result(args)
            if future.done():
                self._waiting.remove((e, check, future))

    def add_event(self, callback: Callable, name: str = None) -> None:
        try:
            func = callback.func
        except AttributeError:
            func = callback

        if not inspect.iscoroutine(func) and not inspect.iscoroutinefunction(func):
            raise ValueError("Event callback must be a coroutine")

        event_name = name or callback.__name__
        self.registered_callbacks[callback] = event_name

        if event_name in self._events:
            self._events[event_name].append(callback)

        else:
            self._events[event_name] = [callback]

    def remove_event(self, callback: Callable) -> bool:
        event_name = self.registered_callbacks.get(callback)

        if event_name is None:
            raise ValueError("Event callback is not a registered event")

        if callback in self._events[event_name]:
            self._events[event_name].remove(callback)
            return True

        return False

    def event(self, name: str = None) -> Callable:
        def decorator(func: Callable) -> Callable:
            self.add_event(func, name)
            return func

        return decorator

    async def wait_for(
        self,
        event: str,
        predicate: Callable[[], bool] = lambda *a: True,
        *,
        timeout=60.0,
    ) -> Tuple[Any]:
        """|coro|


        Waits for an event to be dispatched, then returns the events data

        Parameters
        -----------
        event: :class:`str`
            The event to wait for. Do not include the `event_` prefix
        predicate: Callable[[...], bool]
            A check that is fired when the desired event is dispatched. if the check returns false,
            the waiting will continue until the timeout.
        timeout: :class:`int`
            How long to wait before timing out and raising an error.

        Returns
        --------
        The arguments passed to the event.
        """
        fut = self.loop.create_future()
        tup = (event, predicate, fut)
        self._waiting.append(tup)
        values = await asyncio.wait_for(fut, timeout)
        return values

    def wait_for_ready(self) -> Coroutine[Any, Any, bool]:
        """|coro|

        Waits for the underlying connection to finish startup

        Returns
        --------
        :class:`bool` The state of the underlying flag. This will always be ``True``
        """
        return self._connection.is_ready.wait()

    @id_cache()
    def get_channel(self, name: str) -> Optional[Channel]:
        """Retrieve a channel from the cache.

        Parameters
        -----------
        name: str
            The channel name to retrieve from cache. Returns None if no channel was found.

        Returns
        --------
        :class:`.Channel`
        """
        name = name.lower()

        if name in self._connection._cache:
            # Basically the cache doesn't store channels naturally, instead it stores a channel key
            # With the associated users as a set.
            # We create a Channel here and return it only if the cache has that channel key.

            return Channel(name=name, websocket=self._connection)

    async def part_channels(self, channels: Union[List[str], Tuple[str]]):
        """|coro|

        Part the specified channels.

        Parameters
        ------------
        channels: Union[List[str], Tuple[str]]
            The channels in either a list or tuple form to part.
        """
        await self._connection.part_channels(*channels)

    async def join_channels(self, channels: Union[List[str], Tuple[str]]):
        """|coro|

        Join the specified channels.

        Parameters
        ------------
        channels: Union[List[str], Tuple[str]]
            The channels in either a list or tuple form to join.
        """
        await self._connection.join_channels(*channels)

    @property
    def connected_channels(self) -> List[Channel]:
        """A list of currently connected :class:`.Channel`"""
        return [self.get_channel(x) for x in self._connection._cache.keys()]

    @property
    def events(self):
        """A mapping of events name to coroutine."""
        return self._events

    @property
    def nick(self):
        """The IRC bots nick."""
        return self._http.nick or self._connection.nick

    @property
    def user_id(self):
        """The IRC bot user id."""
        return self._http.user_id or self._connection.user_id

    def create_user(self, user_id: int, user_name: str) -> PartialUser:
        """
        A helper method to create a :class:`twitchio.PartialUser` from a user id and user name.

        Parameters
        -----------
        user_id: :class:`int`
            The id of the user
        user_name: :class:`str`
            The name of the user

        Returns
        --------
        :class:`twitchio.PartialUser`
        """
        return PartialUser(self._http, user_id, user_name)

    @user_cache()
    async def fetch_users(
        self,
        names: List[str] = None,
        ids: List[int] = None,
        token: str = None,
        force=False,
    ) -> List[User]:
        """|coro|

        Fetches users from the helix API

        Parameters
        -----------
        names: Optional[List[:class:`str`]]
            usernames of people to fetch
        ids: Optional[List[:class:`str`]]
            ids of people to fetch
        token: Optional[:class:`str`]
            An optional OAuth token to use instead of the bot OAuth token
        force: :class:`bool`
            whether to force a fetch from the api, or check the cache first. Defaults to False

        Returns
        --------
        List[:class:`twitchio.User`]
        """
        # the forced argument doesnt actually get used here, it gets used by the cache wrapper.
        # But we'll include it in the args here so that sphinx catches it
        assert names or ids
        data = await self._http.get_users(ids, names, token=token)
        return [User(self._http, x) for x in data]

    async def fetch_clips(self, ids: List[str]):
        """|coro|

        Fetches clips by clip id.
        To fetch clips by user id, use :meth:`twitchio.PartialUser.fetch_clips`

        Parameters
        -----------
        ids: List[:class:`str`]
            A list of clip ids

        Returns
        --------
        List[:class:`twitchio.Clip`]
        """
        data = await self._http.get_clips(ids=ids)
        return [models.Clip(self._http, d) for d in data]

    async def fetch_channel(self, broadcaster: str, token: Optional[str] = None):
        """|coro|

        Retrieve channel information from the API.

        .. note::
            This will be deprecated in 3.0. It's recommended to use :func:`~fetch_channels` instead.

        Parameters
        -----------
        broadcaster: str
            The channel name or ID to request from API. Returns empty dict if no channel was found.
        token: Optional[:class:`str`]
            An optional OAuth token to use instead of the bot OAuth token.

        Returns
        --------
        :class:`twitchio.ChannelInfo`
        """

        if not broadcaster.isdigit():
            get_id = await self.fetch_users(names=[broadcaster.lower()])
            if not get_id:
                raise IndexError("Invalid channel name.")
            broadcaster = str(get_id[0].id)
        try:
            data = await self._http.get_channels(broadcaster_id=broadcaster, token=token)

            from .models import ChannelInfo

            return ChannelInfo(self._http, data=data[0])

        except HTTPException:
            raise HTTPException("Incorrect channel ID.")

    async def fetch_channels(self, broadcaster_ids: List[int], token: Optional[str] = None):
        """|coro|

        Retrieve information for up to 100 channels from the API.

        Parameters
        -----------
        broadcaster_ids: List[:class:`int`]
            The channel ids to request from API.
        token: Optional[:class:`str`]
            An optional OAuth token to use instead of the bot OAuth token

        Returns
        --------
        List[:class:`twitchio.ChannelInfo`]
        """
        from .models import ChannelInfo

        data = await self._http.get_channels_new(broadcaster_ids=broadcaster_ids, token=token)
        return [ChannelInfo(self._http, data=d) for d in data]

    async def fetch_videos(
        self,
        ids: List[int] = None,
        game_id: int = None,
        user_id: int = None,
        period=None,
        sort=None,
        type=None,
        language=None,
    ):
        """|coro|

        Fetches videos by id, game id, or user id

        Parameters
        -----------
        ids: Optional[List[:class:`int`]]
            A list of video ids
        game_id: Optional[:class:`int`]
            A game to fetch videos from
        user_id: Optional[:class:`int`]
            A user to fetch videos from. See :meth:`twitchio.PartialUser.fetch_videos`
        period: Optional[:class:`str`]
            The period for which to fetch videos. Valid values are `all`, `day`, `week`, `month`. Defaults to `all`.
            Cannot be used when video id(s) are passed
        sort: :class:`str`
            Sort orders of the videos. Valid values are `time`, `trending`, `views`, Defaults to `time`.
            Cannot be used when video id(s) are passed
        type: Optional[:class:`str`]
            Type of the videos to fetch. Valid values are `upload`, `archive`, `highlight`. Defaults to `all`.
            Cannot be used when video id(s) are passed
        language: Optional[:class:`str`]
            Language of the videos to fetch. Must be an `ISO-639-1 <https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes>`_ two letter code.
            Cannot be used when video id(s) are passed

        Returns
        --------
        List[:class:`twitchio.Video`]
        """
        from .models import Video

        data = await self._http.get_videos(
            ids,
            user_id=user_id,
            game_id=game_id,
            period=period,
            sort=sort,
            type=type,
            language=language,
        )
        return [Video(self._http, x) for x in data]

    async def fetch_cheermotes(self, user_id: int = None):
        """|coro|


        Fetches cheermotes from the twitch API

        Parameters
        -----------
        user_id: Optional[:class:`int`]
            The channel id to fetch from.

        Returns
        --------
        List[:class:`twitchio.CheerEmote`]
        """
        data = await self._http.get_cheermotes(str(user_id) if user_id else None)
        return [models.CheerEmote(self._http, x) for x in data]

    async def fetch_global_emotes(self):
        """|coro|

        Fetches global emotes from the twitch API

        Returns
        --------
        List[:class:`twitchio.GlobalEmote`]
        """
        from .models import GlobalEmote

        data = await self._http.get_global_emotes()
        return [GlobalEmote(self._http, x) for x in data]

    async def fetch_top_games(self) -> List[models.Game]:
        """|coro|

        Fetches the top games from the api

        Returns
        --------
        List[:class:`twitchio.Game`]
        """
        data = await self._http.get_top_games()
        return [models.Game(d) for d in data]

    async def fetch_games(
        self, ids: Optional[List[int]] = None, names: Optional[List[str]] = None, igdb_ids: Optional[List[int]] = None
    ) -> List[models.Game]:
        """|coro|

        Fetches games by id or name.
        At least one id or name must be provided

        Parameters
        -----------
        ids: Optional[List[:class:`int`]]
            An optional list of game ids
        names: Optional[List[:class:`str`]]
            An optional list of game names
        igdb_ids: Optional[List[:class:`int`]]
            An optional list of IGDB game ids

        Returns
        --------
        List[:class:`twitchio.Game`]
        """

        data = await self._http.get_games(ids, names, igdb_ids)
        return [models.Game(d) for d in data]

    async def fetch_tags(self, ids: Optional[List[str]] = None):
        """|coro|

        Fetches stream tags.

        Parameters
        -----------
        ids: Optional[List[:class:`str`]]
            The ids of the tags to fetch

        Returns
        --------
        List[:class:`twitchio.Tag`]
        """
        data = await self._http.get_stream_tags(ids)
        return [models.Tag(x) for x in data]

    async def fetch_streams(
        self,
        user_ids: Optional[List[int]] = None,
        game_ids: Optional[List[int]] = None,
        user_logins: Optional[List[str]] = None,
        languages: Optional[List[str]] = None,
        token: Optional[str] = None,
        type: Literal["all", "live"] = "all",
    ):
        """|coro|

        Fetches live streams from the helix API

        Parameters
        -----------
        user_ids: Optional[List[:class:`int`]]
            user ids of people whose streams to fetch
        game_ids: Optional[List[:class:`int`]]
            game ids of streams to fetch
        user_logins: Optional[List[:class:`str`]]
            user login names of people whose streams to fetch
        languages: Optional[List[:class:`str`]]
            language for the stream(s). ISO 639-1 or two letter code for supported stream language
        token: Optional[:class:`str`]
            An optional OAuth token to use instead of the bot OAuth token
        type: Literal["all", "live"]
            One of ``"all"`` or ``"live"``. Defaults to ``"all"``. Specifies what type of stream to fetch.

        Returns
        --------
        List[:class:`twitchio.Stream`]
        """
        from .models import Stream

        data = await self._http.get_streams(
            game_ids=game_ids,
            user_ids=user_ids,
            user_logins=user_logins,
            languages=languages,
            type_=type,
            token=token,
        )
        return [Stream(self._http, x) for x in data]

    async def fetch_teams(
        self,
        team_name: Optional[str] = None,
        team_id: Optional[str] = None,
    ):
        """|coro|

        Fetches information for a specific Twitch Team.

        Parameters
        -----------
        name: Optional[:class:`str`]
            Team name to fetch
        id: Optional[:class:`str`]
            Team id to fetch

        Returns
        --------
        List[:class:`twitchio.Team`]
        """
        from .models import Team

        assert team_name or team_id
        data = await self._http.get_teams(
            team_name=team_name,
            team_id=team_id,
        )

        return Team(self._http, data[0])

    async def search_categories(self, query: str):
        """|coro|

        Searches twitches categories

        Parameters
        -----------
        query: :class:`str`
            The query to search for

        Returns
        --------
        List[:class:`twitchio.Game`]
        """
        data = await self._http.get_search_categories(query)
        return [models.Game(x) for x in data]

    async def search_channels(self, query: str, *, live_only=False):
        """|coro|

        Searches channels for the given query

        Parameters
        -----------
        query: :class:`str`
            The query to search for
        live_only: :class:`bool`
            Only search live channels. Defaults to False

        Returns
        --------
        List[:class:`twitchio.SearchUser`]
        """
        data = await self._http.get_search_channels(query, live=live_only)
        return [SearchUser(self._http, x) for x in data]

    async def delete_videos(self, token: str, ids: List[int]) -> List[int]:
        """|coro|

        Delete videos from the api. Returns the video ids that were successfully deleted.

        Parameters
        -----------
        token: :class:`str`
            An oauth token with the ``channel:manage:videos`` scope
        ids: List[:class:`int`]
            A list of video ids from the channel of the oauth token to delete

        Returns
        --------
        List[:class:`int`]
        """
        resp = []
        for chunk in [ids[x : x + 3] for x in range(0, len(ids), 3)]:
            resp.append(await self._http.delete_videos(token, chunk))

        return resp

    async def fetch_chatters_colors(self, user_ids: List[int], token: Optional[str] = None):
        """|coro|

        Fetches the color of a chatter.

        Parameters
        -----------
        user_ids: List[:class:`int`]
            List of user ids to fetch the colors for
        token: Optional[:class:`str`]
            An optional user oauth token

        Returns
        --------
        List[:class:`twitchio.ChatterColor`]
        """
        data = await self._http.get_user_chat_color(user_ids, token)
        return [models.ChatterColor(self._http, x) for x in data]

    async def update_chatter_color(self, token: str, user_id: int, color: str):
        """|coro|

        Updates the color of the specified user in the specified channel/broadcaster's chat.

        Parameters
        -----------
        token: :class:`str`
            An oauth token with the ``user:manage:chat_color`` scope.
        user_id: :class:`int`
            The ID of the user whose color is being updated, this must match the user ID in the token.
        color: :class:`str`
            Turbo and Prime users may specify a named color or a Hex color code like #9146FF.
            Please see the Twitch documentation for more information.

        Returns
        --------
        None
        """
        await self._http.put_user_chat_color(token=token, user_id=str(user_id), color=color)

    async def fetch_global_chat_badges(self):
        """|coro|

        Fetches Twitch's list of chat badges, which users may use in any channel's chat room.

        Returns
        --------
        List[:class:`twitchio.ChatBadge`]
        """

        data = await self._http.get_global_chat_badges()
        return [models.ChatBadge(x) for x in data]

    async def fetch_content_classification_labels(self, locale: Optional[str] = None):
        """|coro|

        Fetches information about Twitch content classification labels.

        Parameters
        -----------
        locale: Optional[:class:`str`]
            Locale for the Content Classification Labels.
            You may specify a maximum of 1 locale. Default: “en-US”

        Returns
        --------
        List[:class:`twitchio.ContentClassificationLabel`]
        """
        locale = "en-US" if locale is None else locale
        data = await self._http.get_content_classification_labels(locale)
        return [models.ContentClassificationLabel(x) for x in data]

    async def get_webhook_subscriptions(self):
        """|coro|

        Fetches your current webhook subscriptions. Requires your bot to be logged in with an app access token.

        Returns
        --------
        List[:class:`twitchio.WebhookSubscription`]
        """
        data = await self._http.get_webhook_subs()
        return [models.WebhookSubscription(x) for x in data]

    async def event_token_expired(self):
        """|coro|


        A special event called when the oauth token expires. This is a hook into the http system, it will call this
        when a call to the api fails due to a token expiry. This function should return either a new token, or `None`.
        Returning `None` will cause the client to attempt an automatic token generation.

        .. note::
            This event is a callback hook. It is not a dispatched event. Any errors raised will be passed to the
            :func:`~event_error` event.
        """
        return None

    async def event_mode(self, channel: Channel, user: User, status: str):
        """|coro|


        Event called when a MODE is received from Twitch.

        Parameters
        ------------
        channel: :class:`.Channel`
            Channel object relevant to the MODE event.
        user: :class:`.User`
            User object containing relevant information to the MODE.
        status: str
            The JTV status received by Twitch. Could be either o+ or o-.
            Indicates a moderation promotion/demotion to the :class:`.User`
        """
        pass

    async def event_userstate(self, user: User):
        """|coro|


        Event called when a USERSTATE is received from Twitch.

        Parameters
        ------------
        user: :class:`.User`
            User object containing relevant information to the USERSTATE.
        """
        pass

    async def event_raw_usernotice(self, channel: Channel, tags: dict):
        """|coro|


        Event called when a USERNOTICE is received from Twitch.
        Since USERNOTICE's can be fairly complex and vary, the following sub-events are available:

            :meth:`event_usernotice_subscription` :
            Called when a USERNOTICE Subscription or Re-subscription event is received.

        .. tip::

            For more information on how to handle USERNOTICE's visit:
            https://dev.twitch.tv/docs/irc/tags/#usernotice-twitch-tags


        Parameters
        ------------
        channel: :class:`.Channel`
            Channel object relevant to the USERNOTICE event.
        tags : dict
            A dictionary with the relevant information associated with the USERNOTICE.
            This could vary depending on the event.
        """
        pass

    async def event_usernotice_subscription(self, metadata):
        """|coro|


        Event called when a USERNOTICE subscription or re-subscription event is received from Twitch.

        Parameters
        ------------
        metadata: :class:`NoticeSubscription`
            The object containing various metadata about the subscription event.
            For ease of use, this contains a :class:`User` and :class:`Channel`.

        """
        pass

    async def event_part(self, user: User):
        """|coro|


        Event called when a PART is received from Twitch.

        Parameters
        ------------
        user: :class:`.User`
            User object containing relevant information to the PART.
        """
        pass

    async def event_join(self, channel: Channel, user: User):
        """|coro|


        Event called when a JOIN is received from Twitch.

        Parameters
        ------------
        channel: :class:`.Channel`
            The channel associated with the JOIN.
        user: :class:`.User`
            User object containing relevant information to the JOIN.
        """
        pass

    async def event_message(self, message: Message):
        """|coro|


        Event called when a PRIVMSG is received from Twitch.

        Parameters
        ------------
        message: :class:`.Message`
            Message object containing relevant information.
        """
        pass

    async def event_error(self, error: Exception, data: str = None):
        """|coro|


        Event called when an error occurs while processing data.

        Parameters
        ------------
        error: Exception
            The exception raised.
        data: str
            The raw data received from Twitch. Depending on how this is called, this could be None.

        Example
        ---------
        .. code:: py

            @bot.event()
            async def event_error(error, data):
                traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)
        """
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    async def event_ready(self):
        """|coro|


        Event called when the Bot has logged in and is ready.

        Example
        ---------
        .. code:: py

            @bot.event()
            async def event_ready():
                print(f'Logged into Twitch | {bot.nick}')
        """
        pass

    async def event_reconnect(self):
        """|coro|

        Event called when twitch sends a RECONNECT notice.
        The library will automatically handle reconnecting when such an event is received
        """

    async def event_raw_data(self, data: str):
        """|coro|


        Event called with the raw data received by Twitch.

        Parameters
        ------------
        data: str
            The raw data received from Twitch.

        Example
        ---------
        .. code:: py

            @bot.event()
            async def event_raw_data(data):
                print(data)
        """
        pass

    async def event_channel_joined(self, channel: Channel):
        """|coro|

        Event called when the bot joins a channel.

        Parameters
        ----------
        channel: :class:`.Channel`
            The channel that was joined.
        """
        pass

    async def event_channel_join_failure(self, channel: str):
        """|coro|

        Event called when the bot fails to join a channel.

        Parameters
        ----------
        channel: `str`
            The channel name that was attempted to be joined.
        """
        logger.error(f'The channel "{channel}" was unable to be joined. Check the channel is valid.')

    async def event_raw_notice(self, data: str):
        """|coro|


        Event called with the raw NOTICE data received by Twitch.

        Parameters
        ------------
        data: str
            The raw NOTICE data received from Twitch.

        Example
        ---------
        .. code:: py

            @bot.event()
            async def event_raw_notice(data):
                print(data)
        """
        pass

    async def event_notice(self, message: str, msg_id: Optional[str], channel: Optional[Channel]):
        """|coro|


        Event called with the NOTICE data received by Twitch.

        .. tip::

            For more information on NOTICE msg_ids visit:
            https://dev.twitch.tv/docs/irc/msg-id/

        Parameters
        ------------
        message: :class:`str`
            The message of the NOTICE.
        msg_id: Optional[:class:`str`]
            The msg_id that indicates what the NOTICE type.
        channel: Optional[:class:`~twitchio.Channel`]
            The channel the NOTICE message originated from.

        Example
        ---------
        .. code:: py

            @bot.event()
            async def event_notice(message, msg_id, channel):
                print(message)
        """
        pass



"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
import re
from typing import TYPE_CHECKING

from twitchio import User, PartialUser, Chatter, PartialChatter, Channel, Clip
from .errors import BadArgument

if TYPE_CHECKING:
    from .core import Context


__all__ = (
    "convert_Chatter",
    "convert_Clip",
    "convert_Channel",
    "convert_PartialChatter",
    "convert_PartialUser",
    "convert_User",
)


async def convert_Chatter(ctx: Context, arg: str) -> Chatter:
    """
    Converts the argument into a chatter in the chat. If the chatter is not found, BadArgument is raised.
    """
    arg = arg.lstrip("@")
    resp = [x for x in filter(lambda c: c.name == arg, ctx.chatters or tuple())]
    if not resp:
        raise BadArgument(f"The user '{arg}' was not found in {ctx.channel.name}'s chat.")

    return resp[0]


async def convert_PartialChatter(ctx: Context, arg: str) -> PartialChatter:
    """
    Converts the argument into a chatter in the chat. As opposed to Chatter converter, this will return a PartialChatter regardless of the cache state.
    """
    return PartialChatter(ctx._ws, name=arg.lstrip("@"), channel=ctx.channel, message=None)


async def convert_Clip(ctx: Context, arg: str) -> Clip:
    finder = re.search(r"(https://clips.twitch.tv/)?(?P<slug>.*)", arg)
    if not finder:
        raise RuntimeError(
            "regex failed to match"
        )  # this should never ever raise, but its here to make type checkers happy

    slug = finder.group("slug")
    clips = await ctx.bot.fetch_clips([slug])
    if not clips:
        raise BadArgument(f"Clip '{slug}' was not found")

    return clips[0]


async def convert_User(ctx: Context, arg: str) -> User:
    """
    Similar to convert_Chatter, but fetches from the twitch API instead,
    returning a :class:`twitchio.User` instead of a :class:`twitchio.Chatter`.
    To use this, you most have a valid client id and API token or client secret
    """
    arg = arg.lstrip("@")
    user = await ctx.bot.fetch_users(names=[arg])
    if not user:
        raise BadArgument(f"User '{arg}' was not found.")
    return user[0]


async def convert_PartialUser(ctx: Context, arg: str) -> User:
    """
    This is simply a shorthand to :ref:`~convert_User`, as fetching from the api will return a full user model
    """
    return await convert_User(ctx, arg)


async def convert_Channel(ctx: Context, arg: str) -> Channel:
    if arg not in ctx.bot._connection._cache:
        raise BadArgument(f"Not connected to channel '{arg}'")

    return ctx.bot.get_channel(arg)


_mapping = {
    User: convert_User,
    PartialUser: convert_PartialUser,
    Channel: convert_Channel,
    Chatter: convert_Chatter,
    PartialChatter: convert_PartialChatter,
    Clip: convert_Clip,
}



import datetime

try:
    import ciso8601

    HAS_CISO = True
except (ImportError, ModuleNotFoundError):
    HAS_CISO = False
    import iso8601

__all__ = ("parse_timestamp",)


def parse_timestamp(timestamp: str) -> datetime.datetime:
    """

    Parameters
    ----------
    timestamp: :class:`str`
        The timestamp to be parsed, in an iso8601 format.

    Returns
    -------
    :class:`datetime.datetime`
        The parsed timestamp.

    """
    if HAS_CISO:
        return ciso8601.parse_datetime(timestamp)

    return iso8601.parse_date(timestamp, datetime.timezone.utc)



from __future__ import annotations
import datetime
import hmac
import hashlib
import logging
from enum import Enum
from typing import Any, Dict, TYPE_CHECKING, Optional, Type, Union, Tuple, List, overload
from typing_extensions import Literal

from aiohttp import web

from twitchio import PartialUser, parse_timestamp as _parse_datetime

if TYPE_CHECKING:
    from .server import EventSubClient
    from .websocket import EventSubWSClient

try:
    import ujson as json

    def _loads(s: str) -> dict:
        return json.loads(s)

except ModuleNotFoundError:
    import json

    def _loads(s: str) -> dict:
        return json.loads(s)


logger = logging.getLogger("twitchio.ext.eventsub")


class EmptyObject:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class Subscription:
    __slots__ = "id", "status", "type", "version", "cost", "condition", "transport", "transport_method", "created_at"

    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.status: str = data["status"]
        self.type: str = data["type"]
        self.version = int(data["version"])
        self.cost: int = data["cost"]
        self.condition: Dict[str, str] = data["condition"]
        self.created_at = _parse_datetime(data["created_at"])
        self.transport = EmptyObject()
        self.transport_method: TransportType = getattr(TransportType, data["transport"]["method"])
        self.transport.method: str = data["transport"]["method"]  # type: ignore

        if self.transport_method is TransportType.webhook:
            self.transport.callback: str = data["transport"]["callback"]  # type: ignore
        else:
            self.transport.callback: str = ""  # type: ignore # compatibility
            self.transport.session_id: str = data["transport"]["session_id"]  # type: ignore


class Headers:
    """
    The headers of the inbound EventSub message

    Attributes
    -----------
    message_id: :class:`str`
        The unique ID of the message
    message_retry: :class:`int`
        Unknown
    signature: :class:`str`
        The signature associated with the message
    subscription_type: :class:`str`
        The type of the subscription on the inbound message
    subscription_version: :class:`str`
        The version of the subscription.
    timestamp: :class:`datetime.datetime`
        The timestamp the message was sent at
    """

    def __init__(self, request: web.Request):
        self.message_id: str = request.headers["Twitch-Eventsub-Message-Id"]
        self.message_retry: int = int(request.headers["Twitch-Eventsub-Message-Retry"])
        self.message_type: str = request.headers["Twitch-Eventsub-Message-Type"]
        self.signature: str = request.headers["Twitch-Eventsub-Message-Signature"]
        self.subscription_type: str = request.headers["Twitch-Eventsub-Subscription-Type"]
        self.subscription_version: str = request.headers["Twitch-Eventsub-Subscription-Version"]
        self.timestamp = _parse_datetime(request.headers["Twitch-Eventsub-Message-Timestamp"])
        self._raw_timestamp = request.headers["Twitch-Eventsub-Message-Timestamp"]


class WebsocketHeaders:
    """
    The headers of the inbound Websocket EventSub message

    Attributes
    -----------
    message_id: :class:`str`
        The unique ID of the message
    message_type: :class:`str`
        The type of the message coming through
    message_retry: :class:`int`
        Kept for compatibility with :class:`Headers`
    signature: :class:`str`
        Kept for compatibility with :class:`Headers`
    subscription_type: :class:`str`
        The type of the subscription on the inbound message
    subscription_version: :class:`str`
        The version of the subscription.
    timestamp: :class:`datetime.datetime`
        The timestamp the message was sent at
    """

    def __init__(self, frame: dict):
        meta = frame["metadata"]
        self.message_id: str = meta["message_id"]
        self.timestamp = _parse_datetime(meta["message_timestamp"])
        self.message_type: Literal["notification", "revocation", "reconnect", "session_keepalive"] = meta[
            "message_type"
        ]
        self.message_retry: int = 0  # don't make breaking changes with the Header class
        self.signature: str = ""
        self.subscription_type: Optional[str]
        self.subscription_version: Optional[str]
        if frame["payload"]:
            self.subscription_type = frame["payload"]["subscription"]["type"]
            self.subscription_version = frame["payload"]["subscription"]["version"]
        else:
            self.subscription_type = None
            self.subscription_version = None


class BaseEvent:
    """
    The base of all the event classes

    Attributes
    -----------
    subscription: Optional[:class:`Subscription`]
        The subscription attached to the message. This is only optional when using the websocket eventsub transport
    headers: :class`Headers`
        The headers received with the message
    """

    __slots__ = ("_client", "_raw_data", "subscription", "headers")

    @overload
    def __init__(self, client: EventSubClient, _data: str, request: web.Request):
        ...

    @overload
    def __init__(self, client: EventSubWSClient, _data: dict, request: None):
        ...

    def __init__(
        self, client: Union[EventSubClient, EventSubWSClient], _data: Union[str, dict], request: Optional[web.Request]
    ):
        self._client = client
        self._raw_data = _data

        if isinstance(_data, str):
            data: dict = _loads(_data)
        else:
            data = _data

        self.headers: Union[Headers, WebsocketHeaders]
        self.subscription: Optional[Subscription]

        if request:
            data: dict = _loads(_data)
            self.headers = Headers(request)
            self.subscription = Subscription(data["subscription"])
            self.setup(data)
        else:
            self.headers = WebsocketHeaders(data)
            if data["payload"]:
                self.subscription = Subscription(data["payload"]["subscription"])
            else:
                self.subscription = None
            self.setup(data["payload"])

    def setup(self, data: dict):
        pass

    def verify(self):
        """
        Only used in webhook transport types. Verifies the message is valid
        """
        hmac_message = (self.headers.message_id + self.headers._raw_timestamp + self._raw_data).encode("utf-8")  # type: ignore
        secret = self._client.secret.encode("utf-8")
        digest = hmac.new(secret, msg=hmac_message, digestmod=hashlib.sha256).hexdigest()

        if not hmac.compare_digest(digest, self.headers.signature[7:]):
            logger.warning(f"Recieved a message with an invalid signature, discarding.")
            return web.Response(status=400)

        return web.Response(status=200)


class RevokationEvent(BaseEvent):
    pass


class ChallengeEvent(BaseEvent):
    """
    A challenge event.

    .. note::
        These are only dispatched when using :class:`~twitchio.ext.eventsub.EventSubClient`

    Attributes
    -----------
    challenge: :class`str`
        The challenge received from twitch
    """

    __slots__ = ("challenge",)

    def setup(self, data: dict):
        self.challenge: str = data["challenge"]

    def verify(self):
        hmac_message = (self.headers.message_id + self.headers._raw_timestamp + self._raw_data).encode("utf-8")  # type: ignore
        secret = self._client.secret.encode("utf-8")
        digest = hmac.new(secret, msg=hmac_message, digestmod=hashlib.sha256).hexdigest()

        if not hmac.compare_digest(digest, self.headers.signature[7:]):
            logger.warning(f"Recieved a message with an invalid signature, discarding.")
            return web.Response(status=400)

        return web.Response(status=200, text=self.challenge)


class ReconnectEvent(BaseEvent):
    """
    A reconnect event. Called by twitch when the websocket needs to be disconnected for maintenance or other reasons

    .. note::
        These are only dispatched when using :class:`~twitchio.ext.eventsub.EventSubWSClient`

    Attributes
    -----------
    reconnect_url: :class:`str`
        The URL to reconnect to
    connected_at: :class:`datetime.datetime`
        When the original websocket connected
    """

    __slots__ = ("reconnect_url", "connected_at")

    def __init__(
        self, client: Union[EventSubClient, EventSubWSClient], _data: Union[str, dict], request: Optional[web.Request]
    ):
        # we skip the super init here because reconnect events dont have headers or subscription information

        self._client = client
        self._raw_data = _data

        if isinstance(_data, str):
            data: dict = _loads(_data)
        else:
            data = _data

        self.setup(data["payload"])

    def setup(self, data: dict):
        self.reconnect_url: str = data["session"]["reconnect_url"]
        self.connected_at: datetime.datetime = _parse_datetime(data["session"]["connected_at"])


class KeepAliveEvent(BaseEvent):
    """
    A keep-alive event. Called by twitch when no message has been sent for more than ``keepalive_timeout``

    .. note::
        These are only dispatched when using :class:`~twitchio.ext.eventsub.EventSubWSClient`

    """

    pass


class NotificationEvent(BaseEvent):
    """
    A notification event

    Attributes
    -----------
    data: :class:`models._DataType`
        The data associated with this event
    """

    __slots__ = ("data",)

    def setup(self, _data: dict):
        data: dict = _data["event"]
        typ = self.subscription.type
        if typ not in SubscriptionTypes._type_map:
            raise ValueError(f"Unexpected subscription type '{typ}'")

        self.data: _DataType = SubscriptionTypes._type_map[typ](self._client, data)


def _transform_user(client: EventSubClient, data: dict, field: str) -> PartialUser:
    return client.client.create_user(int(data[field + "_id"]), data[field + "_name"])


class EventData:
    __slots__ = ()


class ChannelBanData(EventData):
    """
    A Ban event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user that was banned
    broadcaster: :class:`twitchio.PartialUser`
        The broadcaster who's channel the ban occurred in
    moderator: :class:`twitchio.PartialUser`
        The moderator responsible for the ban
    reason: :class:`str`
        The reason for the ban
    ends_at: Optional[:class:`datetime.datetime`]
        When the ban ends at. Could be ``None``
    permanant: :class:`bool`
        A typo of ``permanent`` Kept for backwards compatibility
    permanent: :class:`bool`
        Whether the ban is permanent
    """

    __slots__ = "user", "broadcaster", "moderator", "reason", "ends_at", "permenant", "permanent"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.moderator = _transform_user(client, data, "moderator_user")
        self.reason: str = data["reason"]
        self.ends_at: Optional[datetime.datetime] = data["ends_at"] and _parse_datetime(data["ends_at"])
        self.permenant: bool = data["is_permanent"]
        self.permanent = self.permenant  # fix the spelling while keeping backwards compat


class ChannelSubscribeData(EventData):
    """
    A Subscription event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user who subscribed
    broadcaster: :class:`twitchio.PartialUser`
        The channel that was subscribed to
    tier: :class:`int`
        The tier of the subscription
    is_gift: :class:`bool`
        Whether the subscription was a gift or not
    """

    __slots__ = "user", "broadcaster", "tier", "is_gift"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.tier = int(data["tier"])
        self.is_gift: bool = data["is_gift"]


class ChannelSubscriptionEndData(EventData):
    """
    A Subscription End event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user who subscribed
    broadcaster: :class:`twitchio.PartialUser`
        The channel that was subscribed to
    tier: :class:`int`
        The tier of the subscription
    is_gift: :class:`bool`
        Whether the subscription was a gift or not
    """

    __slots__ = "user", "broadcaster", "tier", "is_gift"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.tier = int(data["tier"])
        self.is_gift: bool = data["is_gift"]


class ChannelSubscriptionGiftData(EventData):
    """
    A Subscription Gift event
    Explicitly, the act of giving another user a Subscription.
    Receiving a gift-subscription uses ChannelSubscribeData above, with is_gift is ``True``

    Attributes
    -----------
    is_anonymous: :class:`bool`
        Whether the gift sub was anonymous
    user: Optional[:class:`twitchio.PartialUser`]
        The user that gifted subs. Will be ``None`` if ``is_anonymous`` is ``True``
    broadcaster: :class:`twitchio.PartialUser`
        The channel that was subscribed to
    tier: :class:`int`
        The tier of the subscription
    total: :class:`int`
        The total number of subs gifted by a user at once
    cumulative_total: Optional[:class:`int`]
        The total number of subs gifted by a user overall. Will be ``None`` if ``is_anonymous`` is ``True``
    """

    __slots__ = "is_anonymous", "user", "broadcaster", "tier", "total", "cumulative_total"

    def __init__(self, client: EventSubClient, data: dict):
        self.is_anonymous: bool = data["is_anonymous"]
        self.user: Optional[PartialUser] = None if self.is_anonymous else _transform_user(client, data, "user")
        self.broadcaster: Optional[PartialUser] = _transform_user(client, data, "broadcaster_user")
        self.tier = int(data["tier"])
        self.total = int(data["total"])
        self.cumulative_total: Optional[int] = None if self.is_anonymous else int(data["cumulative_total"])


class ChannelSubscriptionMessageData(EventData):
    """
    A Subscription Message event.
    A combination of resubscriptions + the messages users type as part of the resub.

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user who subscribed
    broadcaster: :class:`twitchio.PartialUser`
        The channel that was subscribed to
    tier: :class:`int`
        The tier of the subscription
    message: :class:`str`
        The user's resubscription message
    emote_data: :class:`list`
        emote data within the user's resubscription message. Not the emotes themselves
    cumulative_months: :class:`int`
        The total number of months a user has subscribed to the channel
    streak: Optional[:class:`int`]
        The total number of months subscribed in a row. ``None`` if the user declines to share it.
    duration: :class:`int`
        The length of the subscription. Typically 1, but some users may buy subscriptions for several months.
    """

    __slots__ = "user", "broadcaster", "tier", "message", "emote_data", "cumulative_months", "streak", "duration"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.tier = int(data["tier"])
        self.message: str = data["message"]["text"]
        self.emote_data: List[Dict] = data["message"].get("emotes", [])
        self.cumulative_months: int = data["cumulative_months"]
        self.streak: Optional[int] = data["streak_months"]
        self.duration: int = data["duration_months"]


class ChannelCheerData(EventData):
    """
    A Cheer event

    Attributes
    ----------
    is_anonymous: :class:`bool`
        Whether the cheer was anonymous
    user: Optional[:class:`twitchio.PartialUser`]
        The user that cheered. Will be ``None`` if ``is_anonymous`` is ``True``
    broadcaster: :class:`twitchio.PartialUser`
        The channel the cheer happened on
    message: :class:`str`
        The message sent along with the bits
    bits: :class:`int`
        The amount of bits sent
    """

    __slots__ = "user", "broadcaster", "is_anonymous", "message", "bits"

    def __init__(self, client: EventSubClient, data: dict):
        self.is_anonymous: bool = data["is_anonymous"]
        self.user: Optional[PartialUser] = _transform_user(client, data, "user") if not self.is_anonymous else None
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.message: str = data["message"]
        self.bits = int(data["bits"])


class ChannelUpdateData(EventData):
    """
    A Channel Update event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel that was updated
    title: :class:`str`
        The title of the stream
    language: :class:`str`
        The language of the channel
    category_id: :class:`str`
        The category the stream is in
    category_name: :class:`str`
        The category the stream is in
    is_mature: :class:`bool`
        Whether the channel is marked as mature by the broadcaster
    """

    __slots__ = "broadcaster", "title", "language", "category_id", "category_name", "is_mature"

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.title: str = data["title"]
        self.language: str = data["language"]
        self.category_id: str = data["category_id"]
        self.category_name: str = data["category_name"]
        self.is_mature: bool = data["is_mature"] == "true"


class ChannelUnbanData(EventData):
    """
    A Channel Unban event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user that was unbanned
    broadcaster: :class:`twitchio.PartialUser`
        The channel the unban occurred in
    moderator: :class`twitchio.PartialUser`
        The moderator that preformed the unban
    """

    __slots__ = "user", "broadcaster", "moderator"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.moderator = _transform_user(client, data, "moderator_user")


class ChannelFollowData(EventData):
    """
    A Follow event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user that followed
    broadcaster: :class:`twitchio.PartialUser`
        The channel that was followed
    followed_at: :class:`datetime.datetime`
        When the follow occurred
    """

    __slots__ = "user", "broadcaster", "followed_at"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.followed_at = _parse_datetime(data["followed_at"])


class ChannelRaidData(EventData):
    """
    A Raid event

    Attributes
    -----------
    raider: :class:`twitchio.PartialUser`
        The person initiating the raid
    reciever: :class:`twitchio.PartialUser`
        The person recieving the raid
    viewer_count: :class:`int`
        The amount of people raiding
    """

    __slots__ = "raider", "reciever", "viewer_count"

    def __init__(self, client: EventSubClient, data: dict):
        self.raider = _transform_user(client, data, "from_broadcaster_user")
        self.reciever = _transform_user(client, data, "to_broadcaster_user")
        self.viewer_count: int = data["viewers"]


class ChannelModeratorAddRemoveData(EventData):
    """
    A Moderator Add/Remove event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user being added or removed from the moderator status
    broadcaster: :class:`twitchio.PartialUser`
        The channel that is having a moderator added/removed
    """

    __slots__ = "broadcaster", "user"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.broadcaster = _transform_user(client, data, "broadcaster_user")


class CustomReward:
    """
    A Custom Reward

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel that has this reward
    id: :class:`str`
        The ID of the reward
    title: :class:`str`
        The title of the reward
    cost: :class:`int`
        The cost of the reward in Channel Points
    prompt: :class:`str`
        The prompt of the reward
    enabled: Optional[:class:`bool`]
        Whether or not the reward is enabled. Will be `None` for Redemption events.
    paused: Optional[:class:`bool`]
        Whether or not the reward is paused. Will be `None` for Redemption events.
    in_stock: Optional[:class:`bool`]
        Whether or not the reward is in stock. Will be `None` for Redemption events.
    cooldown_until: Optional[:class:`datetime.datetime`]
        How long until the reward is off cooldown and can be redeemed again. Will be `None` for Redemption events.
    input_required: Optional[:class:`bool`]
        Whether or not the reward requires an input. Will be `None` for Redemption events.
    redemptions_skip_queue: Optional[:class:`bool`]
        Whether or not redemptions for this reward skips the queue. Will be `None` for Redemption events.
    redemptions_current_stream: Optional[:class:`int`]
        How many redemptions of this reward have been redeemed for this stream. Will be `None` for Redemption events.
    max_per_stream: Tuple[:class:`bool`, :class:`int`]
        Whether or not a per-stream redemption limit is in place, and if so, the maximum number of redemptions allowed
        per stream. Will be `None` for Redemption events.
    max_per_user_per_stream: Tuple[:class:`bool`, :class:`int`]
        Whether or not a per-user-per-stream redemption limit is in place, and if so, the maximum number of redemptions
        allowed per user per stream. Will be `None` for Redemption events.
    cooldown: Tuple[:class:`bool`, :class:`int`]
        Whether or not a global cooldown is in place, and if so, the number of seconds until the reward can be redeemed
        again. Will be `None` for Redemption events.
    background_color: Optional[:class:`str`]
        Hexadecimal color code for the background of the reward.
    image: Optional[:class:`str`]
        Image URL for the reward.
    """

    __slots__ = (
        "broadcaster",
        "id",
        "title",
        "cost",
        "prompt",
        "enabled",
        "paused",
        "in_stock",
        "cooldown_until",
        "input_required",
        "redemptions_skip_queue",
        "redemptions_current_stream",
        "max_per_stream",
        "max_per_user_stream",
        "cooldown",
        "background_color",
        "image",
    )

    def __init__(self, data, broadcaster):
        self.broadcaster: PartialUser = broadcaster

        self.id: str = data["id"]

        self.title: str = data["title"]
        self.cost: int = data["cost"]
        self.prompt: str = data["prompt"]

        self.enabled: Optional[bool] = data.get("is_enabled", None)
        self.paused: Optional[bool] = data.get("is_paused", None)
        self.in_stock: Optional[bool] = data.get("is_in_stock", None)

        self.cooldown_until: Optional[datetime.datetime] = (
            _parse_datetime(data["cooldown_expires_at"]) if data.get("cooldown_expires_at", None) else None
        )

        self.input_required: Optional[bool] = data.get("is_user_input_required", None)
        self.redemptions_skip_queue: Optional[bool] = data.get("should_redemptions_skip_request_queue", None)
        self.redemptions_current_stream: Optional[bool] = data.get("redemptions_redeemed_current_stream", None)

        self.max_per_stream: Tuple[Optional[bool], Optional[int]] = (
            data.get("max_per_stream", {}).get("is_enabled"),
            data.get("max_per_stream", {}).get("value"),
        )
        self.max_per_user_stream: Tuple[Optional[bool], Optional[int]] = (
            data.get("max_per_user_per_stream", {}).get("is_enabled"),
            data.get("max_per_user_per_stream", {}).get("value"),
        )
        self.cooldown: Tuple[Optional[bool], Optional[int]] = (
            data.get("global_cooldown", {}).get("is_enabled"),
            data.get("global_cooldown", {}).get("seconds"),
        )

        self.background_color: Optional[str] = data.get("background_color", None)
        self.image: Optional[str] = data.get("image", data.get("default_image", {})).get("url_1x", None)


class CustomRewardAddUpdateRemoveData(EventData):
    """
    A Custom Reward Add/Update/Remove event

    Attributes
    -----------
    id: :class:`str`
        The ID of the custom reward
    broadcaster: :class:`twitchio.PartialUser`
        The channel the custom reward was modified in
    reward: :class:`CustomReward`
        The reward object
    """

    __slots__ = "reward", "broadcaster", "id"

    def __init__(self, client: EventSubClient, data: dict):
        self.id: str = data["id"]
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.reward = CustomReward(data, self.broadcaster)


class CustomRewardRedemptionAddUpdateData(EventData):
    """
    A Custom Reward Redemption event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel the redemption occurred in
    user: :class:`twitchio.PartialUser`
        The user that redeemed the reward
    id: :class:`str`
        The ID of the redemption
    input: :class:`str`
        The user input, if present. This will be an empty string if it is not present
    status: :class:`str`
        One of "unknown", "unfulfilled", "fulfilled", or "cancelled"
    redeemed_at: :class:`datetime.datetime`
        When the reward was redeemed at
    reward: :class:`CustomReward`
        The reward object
    """

    __slots__ = "broadcaster", "id", "user", "input", "status", "reward", "redeemed_at"

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.user = _transform_user(client, data, "user")
        self.id: str = data["id"]
        self.input: str = data["user_input"]
        self.status: Literal["unknown", "unfulfilled", "fulfilled", "cancelled"] = data["status"]
        self.redeemed_at = _parse_datetime(data["redeemed_at"])
        self.reward = CustomReward(data["reward"], self.broadcaster)


class HypeTrainContributor:
    """
    A Contributor to a Hype Train

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user
    type: :class:`str`
        One of "bits, "subscription" or "other". The way they contributed to the hype train
    total: :class:`int`
        How many points they've contributed to the Hype Train
    """

    __slots__ = "user", "type", "total"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.type: Literal["bits", "subscription", "other"] = data["type"]  # one of bits, subscription
        self.total: int = data["total"]


class HypeTrainBeginProgressData(EventData):
    """
    A Hype Train Begin/Progress event

    Attributes
    -----------

    broadcaster: :class:`twitchio.PartialUser`
        The channel the Hype Train occurred in
    total_points: :class:`int`
        The total amounts of points in the Hype Train
    progress: :class:`int`
        The progress of the Hype Train towards the next level
    goal: :class:`int`
        The goal to reach the next level
    started: :class:`datetime.datetime`
        When the Hype Train started
    expires: :class:`datetime.datetime`
        When the Hype Train ends
    top_contributions: List[:class:`HypeTrainContributor`]
        The top contributions of the Hype Train
    last_contribution: :class:`HypeTrainContributor`
        The last contributor to the Hype Train
    level: :class:`int`
        The current level of the Hype Train
    """

    __slots__ = (
        "broadcaster",
        "total_points",
        "progress",
        "goal",
        "top_contributions",
        "last_contribution",
        "started",
        "expires",
        "level",
    )

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.total_points: int = data["total"]
        self.progress: int = data["progress"]
        self.goal: int = data["goal"]
        self.started = _parse_datetime(data["started_at"])
        self.expires = _parse_datetime(data["expires_at"])
        self.top_contributions = [HypeTrainContributor(client, d) for d in data["top_contributions"]]
        self.last_contribution = HypeTrainContributor(client, data["last_contribution"])
        self.level: int = data["level"]


class HypeTrainEndData(EventData):
    """
    A Hype Train End event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel the Hype Train occurred in
    total_points: :class:`int`
        The total amounts of points in the Hype Train
    level: :class:`int`
        The level the hype train reached
    started: :class:`datetime.datetime`
        When the Hype Train started
    top_contributions: List[:class:`HypeTrainContributor`]
        The top contributions of the Hype Train
    cooldown_ends_at: :class:`datetime.datetime`
        When another Hype Train can begin
    """

    __slots__ = "broadcaster", "level", "total_points", "top_contributions", "started", "ended", "cooldown_ends_at"

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.total_points: int = data["total"]
        self.level: int = data["level"]
        self.started = _parse_datetime(data["started_at"])
        self.ended = _parse_datetime(data["ended_at"])
        self.cooldown_ends_at = _parse_datetime(data["cooldown_ends_at"])
        self.top_contributions = [HypeTrainContributor(client, d) for d in data["top_contributions"]]


class PollChoice:
    """
    A Poll Choice

    Attributes
    -----------
    choice_id: :class:`str`
        The ID of the choice
    title: :class:`str`
        The title of the choice
    bits_votes: :class:`int`
        How many votes were cast using Bits

        .. warning::

            Twitch have removed support for voting with bits.
            This will return as 0

    channel_points_votes: :class:`int`
        How many votes were cast using Channel Points
    votes: :class:`int`
        The total number of votes, including votes cast using Bits and Channel Points
    """

    __slots__ = "choice_id", "title", "bits_votes", "channel_points_votes", "votes"

    def __init__(self, data):
        self.choice_id: str = data["id"]
        self.title: str = data["title"]
        self.bits_votes: int = data.get("bits_votes", 0)
        self.channel_points_votes: int = data.get("channel_points_votes", 0)
        self.votes: int = data.get("votes", 0)


class BitsVoting:
    """
    Information on voting on a poll with Bits

    Attributes
    -----------
    is_enabled: :class:`bool`
        Whether users can use Bits to vote on the poll
    amount_per_vote: :class:`int`
        How many Bits are required to cast an extra vote

        .. warning::

            Twitch have removed support for voting with bits.
            This will return as False and 0 respectively

    """

    __slots__ = "is_enabled", "amount_per_vote"

    def __init__(self, data):
        self.is_enabled: bool = data["is_enabled"]
        self.amount_per_vote: int = data["amount_per_vote"]


class ChannelPointsVoting:
    """
    Information on voting on a poll with Channel Points

    Attributes
    -----------
    is_enabled: :class:`bool`
        Whether users can use Channel Points to vote on the poll
    amount_per_vote: :class:`int`
        How many Channel Points are required to cast an extra vote
    """

    __slots__ = "is_enabled", "amount_per_vote"

    def __init__(self, data):
        self.is_enabled: bool = data["is_enabled"]
        self.amount_per_vote: int = data["amount_per_vote"]


class PollStatus(Enum):
    """
    The status of a poll.

    ACTIVE: Poll is currently in progress.
    COMPLETED: Poll has reached its `ended_at` time.
    TERMINATED: Poll has been manually terminated before its `ended_at` time.
    ARCHIVED: Poll is no longer visible on the channel.
    MODERATED: Poll is no longer visible to any user on Twitch.
    INVALID: Something went wrong determining the state.
    """

    ACTIVE = "active"
    COMPLETED = "completed"
    TERMINATED = "terminated"
    ARCHIVED = "archived"
    MODERATED = "moderated"
    INVALID = "invalid"


class PollBeginProgressData(EventData):
    """
    A Poll Begin/Progress event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel the poll occured in
    poll_id: :class:`str`
        The ID of the poll
    title: :class:`str`
        The title of the poll
    choices: List[:class:`PollChoice`]
        The choices in the poll
    bits_voting: :class:`BitsVoting`
        Information on voting on the poll with Bits

        .. warning::

            Twitch have removed support for voting with bits.

    channel_points_voting: :class:`ChannelPointsVoting`
        Information on voting on the poll with Channel Points
    started_at: :class:`datetime.datetime`
        When the poll started
    ends_at: :class:`datetime.datetime`
        When the poll is set to end
    ...
    """

    __slots__ = (
        "broadcaster",
        "poll_id",
        "title",
        "choices",
        "bits_voting",
        "channel_points_voting",
        "started_at",
        "ends_at",
    )

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.poll_id: str = data["id"]
        self.title: str = data["title"]
        self.choices = [PollChoice(c) for c in data["choices"]]
        self.bits_voting = BitsVoting(data["bits_voting"])
        self.channel_points_voting = ChannelPointsVoting(data["channel_points_voting"])
        self.started_at = _parse_datetime(data["started_at"])
        self.ends_at = _parse_datetime(data["ends_at"])


class PollEndData(EventData):
    """
    A Poll End event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel the poll occured in
    poll_id: :class:`str`
        The ID of the poll
    title: :class:`str`
        The title of the poll
    choices: List[:class:`PollChoice`]
        The choices in the poll
    bits_voting: :class:`BitsVoting`
        Information on voting on the poll with Bits

        .. warning::

            Twitch have removed support for voting with bits.

    channel_points_voting: :class:`ChannelPointsVoting`
        Information on voting on the poll with Channel Points
    status: :class:`PollStatus`
        How the poll ended
    started_at: :class:`datetime.datetime`
        When the poll started
    ended_at: :class:`datetime.datetime`
        When the poll is set to end
    """

    __slots__ = (
        "broadcaster",
        "poll_id",
        "title",
        "choices",
        "bits_voting",
        "channel_points_voting",
        "status",
        "started_at",
        "ended_at",
    )

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.poll_id: str = data["id"]
        self.title: str = data["title"]
        self.choices = [PollChoice(c) for c in data["choices"]]
        self.bits_voting = BitsVoting(data["bits_voting"])
        self.channel_points_voting = ChannelPointsVoting(data["channel_points_voting"])
        self.status = PollStatus(data["status"].lower())
        self.started_at = _parse_datetime(data["started_at"])
        self.ended_at = _parse_datetime(data["ended_at"])


class Predictor:
    """
    A Predictor

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user who predicted an outcome
    channel_points_used: :class:`int`
        How many Channel Points the user used to predict this outcome
    channel_points_won: :class:`int`
        How many Channel Points was distributed to the user.
        Will be `None` if the Prediction is unresolved, cancelled (refunded), or the user predicted the losing outcome.
    """

    __slots__ = "user", "channel_points_used", "channel_points_won"

    def __init__(self, client: EventSubClient, data: dict):
        self.user = _transform_user(client, data, "user")
        self.channel_points_used: int = data["channel_points_used"]
        self.channel_points_won: int = data["channel_points_won"]


class PredictionOutcome:
    """
    A Prediction Outcome

    Attributes
    -----------
    outcome_id: :class:`str`
        The ID of the outcome
    title: :class:`str`
        The title of the outcome
    channel_points: :class:`int`
        The amount of Channel Points that have been bet for this outcome
    color: :class:`str`
        The color of the outcome. Can be `blue` or `pink`
    users: :class:`int`
        The number of users who predicted the outcome
    top_predictors: List[:class:`Predictor`]
        The top predictors of the outcome
    """

    __slots__ = "outcome_id", "title", "channel_points", "color", "users", "top_predictors"

    def __init__(self, client: EventSubClient, data: dict):
        self.outcome_id: str = data["id"]
        self.title: str = data["title"]
        self.channel_points: int = data.get("channel_points", 0)
        self.color: str = data["color"]
        self.users: int = data.get("users", 0)
        self.top_predictors = [Predictor(client, x) for x in data.get("top_predictors", [])]

    @property
    def colour(self) -> str:
        """The colour of the prediction. Alias to color."""
        return self.color


class PredictionStatus(Enum):
    """
    The status of a Prediction.

    ACTIVE: Prediction is active and viewers can make predictions.
    LOCKED: Prediction has been locked and viewers can no longer make predictions.
    RESOLVED: A winning outcome has been chosen and the Channel Points have been distributed to the users who guessed the correct outcome.
    CANCELED: Prediction has been canceled and the Channel Points have been refunded to participants.
    """

    ACTIVE = "active"
    LOCKED = "locked"
    RESOLVED = "resolved"
    CANCELED = "canceled"


class PredictionBeginProgressData(EventData):
    """
    A Prediction Begin/Progress event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel the prediction occured in
    prediction_id: :class:`str`
        The ID of the prediction
    title: :class:`str`
        The title of the prediction
    outcomes: List[:class:`PredictionOutcome`]
        The outcomes for the prediction
    started_at: :class:`datetime.datetime`
        When the prediction started
    locks_at: :class:`datetime.datetime`
        When the prediction is set to be locked
    """

    __slots__ = "broadcaster", "prediction_id", "title", "outcomes", "started_at", "locks_at"

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.prediction_id: str = data["id"]
        self.title: str = data["title"]
        self.outcomes = [PredictionOutcome(client, x) for x in data["outcomes"]]
        self.started_at = _parse_datetime(data["started_at"])
        self.locks_at = _parse_datetime(data["locks_at"])


class PredictionLockData(EventData):
    """
    A Prediction Begin/Progress event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel the prediction occured in
    prediction_id: :class:`str`
        The ID of the prediction
    title: :class:`str`
        The title of the prediction
    outcomes: List[:class:`PredictionOutcome`]
        The outcomes for the prediction
    started_at: :class:`datetime.datetime`
        When the prediction started
    locked_at: :class:`datetime.datetime`
        When the prediction was locked
    """

    __slots__ = "broadcaster", "prediction_id", "title", "outcomes", "started_at", "locked_at"

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.prediction_id: str = data["id"]
        self.title: str = data["title"]
        self.outcomes = [PredictionOutcome(client, x) for x in data["outcomes"]]
        self.started_at = _parse_datetime(data["started_at"])
        self.locked_at = _parse_datetime(data["locked_at"])


class PredictionEndData(EventData):
    """
    A Prediction Begin/Progress event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel the prediction occured in
    prediction_id: :class:`str`
        The ID of the prediction
    title: :class:`str`
        The title of the prediction
    winning_outcome_id: :class:`str`
        The ID of the outcome that won
    outcomes: List[:class:`PredictionOutcome`]
        The outcomes for the prediction
    status: :class:`PredictionStatus`
        How the prediction ended
    started_at: :class:`datetime.datetime`
        When the prediction started
    ended_at: :class:`datetime.datetime`
        When the prediction ended
    """

    __slots__ = (
        "broadcaster",
        "prediction_id",
        "title",
        "winning_outcome_id",
        "outcomes",
        "status",
        "started_at",
        "ended_at",
    )

    def __init__(self, client: EventSubClient, data: dict):
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.prediction_id: str = data["id"]
        self.title: str = data["title"]
        self.winning_outcome_id: str = data["winning_outcome_id"]
        self.outcomes = [PredictionOutcome(client, x) for x in data["outcomes"]]
        self.status = PredictionStatus(data["status"].lower())
        self.started_at = _parse_datetime(data["started_at"])
        self.ended_at = _parse_datetime(data["ended_at"])


class StreamOnlineData(EventData):
    """
    A Stream Start event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel that went live
    id: :class:`str`
        Some sort of ID for the stream
    type: :class:`str`
        One of "live", "playlist", "watch_party", "premier", or "rerun". The type of live event.
    started_at: :class:`datetime.datetime`
    """

    __slots__ = "broadcaster", "id", "type", "started_at"

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.id: str = data["id"]
        self.type: Literal["live", "playlist", "watch_party", "premier", "rerun"] = data["type"]
        self.started_at = _parse_datetime(data["started_at"])


class StreamOfflineData(EventData):
    """
    A Stream End event

    Attributes
    -----------
    broadcaster: :class:`twitchio.PartialUser`
        The channel that stopped streaming
    """

    __slots__ = ("broadcaster",)

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster = _transform_user(client, data, "broadcaster_user")


class UserAuthorizationGrantedData(EventData):
    """
    An Authorization Granted event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user that has granted authorization for your app
    client_id: :class:`str`
        The client id of the app that had its authorization granted
    """

    __slots__ = "client_id", "user"

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.user = _transform_user(client, data, "user")
        self.client_id: str = data["client_id"]


class UserAuthorizationRevokedData(EventData):
    """
    An Authorization Revokation event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user that has revoked authorization for your app
    client_id: :class:`str`
        The client id of the app that had its authorization revoked
    """

    __slots__ = "client_id", "user"

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.user = _transform_user(client, data, "user")
        self.client_id: str = data["client_id"]


class UserUpdateData(EventData):
    """
    A User Update event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The user that was updated
    email: Optional[:class:`str`]
        The users email, if you have permission to read this information
    description: :class:`str`
        The channels description (displayed as ``bio``)
    """

    __slots__ = "user", "email", "description"

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.user = _transform_user(client, data, "user")
        self.email: Optional[str] = data["email"]
        self.description: str = data["description"]


class ChannelGoalBeginProgressData(EventData):
    """
    A goal begin event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The broadcaster that started the goal
    id : :class:`str`
        The ID of the goal event
    type: :class:`str`
        The goal type
    description: :class:`str`
        The goal description
    current_amount: :class:`int`
        The goal current amount
    target_amount: :class:`int`
        The goal target amount
    started_at: :class:`datetime.datetime`
        The datetime the goal was started
    """

    __slots__ = "user", "id", "type", "description", "current_amount", "target_amount", "started_at"

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.user = _transform_user(client, data, "broadcaster_user")
        self.id: str = data["id"]
        self.type: str = data["type"]
        self.description: str = data["description"]
        self.current_amount: int = data["current_amount"]
        self.target_amount: int = data["target_amount"]
        self.started_at: datetime.datetime = _parse_datetime(data["started_at"])


class ChannelGoalEndData(EventData):
    """
    A goal end event

    Attributes
    -----------
    user: :class:`twitchio.PartialUser`
        The broadcaster that ended the goal
    id : :class:`str`
        The ID of the goal event
    type: :class:`str`
        The goal type
    description: :class:`str`
        The goal description
    is_achieved: :class:`bool`
        Whether the goal is achieved
    current_amount: :class:`int`
        The goal current amount
    target_amount: :class:`int`
        The goal target amount
    started_at: :class:`datetime.datetime`
        The datetime the goal was started
    ended_at: :class:`datetime.datetime`
        The datetime the goal was ended
    """

    __slots__ = (
        "user",
        "id",
        "type",
        "description",
        "current_amount",
        "target_amount",
        "started_at",
        "is_achieved",
        "ended_at",
    )

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.user = _transform_user(client, data, "broadcaster_user")
        self.id: str = data["id"]
        self.type: str = data["type"]
        self.description: str = data["description"]
        self.is_achieved: bool = data["is_achieved"]
        self.current_amount: int = data["current_amount"]
        self.target_amount: int = data["target_amount"]
        self.started_at: datetime.datetime = _parse_datetime(data["started_at"])
        self.ended_at: datetime.datetime = _parse_datetime(data["ended_at"])


class ChannelShieldModeBeginData(EventData):
    """
    Represents a Shield Mode activation status.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster whose Shield Mode status was updated.
    moderator: :class:`~twitchio.PartialUser`
        The moderator that updated the Shield Mode staus.
    started_at: :class:`datetime.datetime`
        The UTC datetime of when Shield Mode was last activated.
    """

    __slots__ = ("broadcaster", "moderator", "started_at")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.moderator: PartialUser = _transform_user(client, data, "moderator_user")
        self.started_at: datetime.datetime = _parse_datetime(data["started_at"])


class ChannelShieldModeEndData(EventData):
    """
    Represents a Shield Mode activation status.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster whose Shield Mode status was updated.
    moderator: :class:`~twitchio.PartialUser`
        The moderator that updated the Shield Mode staus.
    ended_at: :class:`datetime.datetime`
        The UTC datetime of when Shield Mode was last deactivated.
    """

    __slots__ = ("broadcaster", "moderator", "ended_at")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.moderator: PartialUser = _transform_user(client, data, "moderator_user")
        self.ended_at: datetime.datetime = _parse_datetime(data["ended_at"])


class ChannelShoutoutCreateData(EventData):
    """
    Represents a Shoutout event being sent.

    Requires the ``moderator:read:shoutouts`` or ``moderator:manage:shoutouts`` scope.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster from who sent the shoutout event.
    moderator: :class:`~twitchio.PartialUser`
        The moderator who sent the shoutout event.
    to_broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster who the shoutout was sent to.
    started_at: :class:`datetime.datetime`
        The datetime the shoutout was sent.
    viewer_count: :class:`int`
        The viewer count at the time of the shoutout
    cooldown_ends_at: :class:`datetime.datetime`
        The datetime the broadcaster can send another shoutout.
    target_cooldown_ends_at: :class:`datetime.datetime`
        The datetime the broadcaster can send another shoutout to the same broadcaster.
    """

    __slots__ = (
        "broadcaster",
        "moderator",
        "to_broadcaster",
        "started_at",
        "viewer_count",
        "cooldown_ends_at",
        "target_cooldown_ends_at",
    )

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.moderator: PartialUser = _transform_user(client, data, "moderator_user")
        self.to_broadcaster: PartialUser = _transform_user(client, data, "to_broadcaster_user")
        self.started_at: datetime.datetime = _parse_datetime(data["started_at"])
        self.viewer_count: int = data["viewer_count"]
        self.cooldown_ends_at: datetime.datetime = _parse_datetime(data["cooldown_ends_at"])
        self.target_cooldown_ends_at: datetime.datetime = _parse_datetime(data["target_cooldown_ends_at"])


class ChannelShoutoutReceiveData(EventData):
    """
    Represents a Shoutout event being received.

    Requires the ``moderator:read:shoutouts`` or ``moderator:manage:shoutouts`` scope.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster receiving shoutout event.
    from_broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster who sent the shoutout.
    started_at: :class:`datetime.datetime`
        The datetime the shoutout was sent.
    viewer_count: :class:`int`
        The viewer count at the time of the shoutout
    """

    __slots__ = ("broadcaster", "from_broadcaster", "started_at", "viewer_count")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.from_broadcaster: PartialUser = _transform_user(client, data, "to_broadcaster_user")
        self.started_at: datetime.datetime = _parse_datetime(data["started_at"])
        self.viewer_count: int = data["viewer_count"]


class ChannelCharityDonationData(EventData):
    """
    Represents a donation towards a charity campaign.

    Requires the ``channel:read:charity`` scope.

    Attributes
    -----------
    id: :class:`str`
        The ID of the event.
    campaign_id: :class:`str`
        The ID of the running charity campaign.
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster running the campaign.
    user: :class:`~twitchio.PartialUser`
        The user who donated.
    charity_name: :class:`str`
        The name of the charity.
    charity_description: :class:`str`
        The description of the charity.
    charity_logo: :class:`str`
        The logo of the charity.
    charity_website: :class:`str`
        The websiet of the charity.
    donation_value: :class:`int`
        The amount of money being donated.
    donation_decimal_places: :class:`int`
        The decimal places to put into the :attr`~.donation_amount`.
    donation_currency: :class:`str`
        The currency that was donated (ex. ``USD``, ``GBP``, ``EUR``)
    """

    __slots__ = (
        "id",
        "campaign_id",
        "broadcaster",
        "user",
        "charity_name",
        "charity_description",
        "charity_logo",
        "charity_website",
        "donation_value",
        "donation_decimal_places",
        "donation_currency",
    )

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.id: str = data["id"]
        self.campaign_id: str = data["campaign_id"]
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster")
        self.user: PartialUser = _transform_user(client, data, "user")
        self.charity_name: str = data["charity_name"]
        self.charity_description: str = data["charity_description"]
        self.charity_logo: str = data["charity_logo"]
        self.charity_website: str = data["charity_website"]
        self.donation_value: int = data["amount"]["value"]
        self.donation_currency: str = data["amount"]["currency"]
        self.donation_decimal_places: int = data["amount"]["decimal_places"]


class ChannelUnbanRequestCreateData(EventData):
    """
    Represents an unban request created by a user.

    Attributes
    -----------
    id: :class:`str`
        The ID of the ban request.
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster from which the user was banned.
    user: :class:`~twitchio.PartialUser`
        The user that was banned.
    text: :class:`str`
        The unban request text the user submitted.
    created_at: :class:`datetime.datetime`
        When the user submitted the request.
    """

    __slots__ = ("id", "broadcaster", "user", "text", "created_at")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.id: str = data["id"]
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster")
        self.user: PartialUser = _transform_user(client, data, "user")
        self.text: str = data["text"]
        self.created_at: datetime.datetime = _parse_datetime(data["created_at"])


class ChannelUnbanRequestResolveData(EventData):
    """
    Represents an unban request that has been resolved by a moderator.

    Attributes
    -----------
    id: :class:`str`
        The ID of the ban request.
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster from which the user was banned.
    user: :class:`~twitchio.PartialUser`
        The user that was banned.
    moderator: :class:`~twitchio.PartialUser`
        The moderator that handled this unban request.
    resolution_text: :class:`str`
        The reasoning provided by the moderator.
    status: :class:`str`
        The resolution. either `accepted` or `denied`.
    """

    __slots__ = ("id", "broadcaster", "user", "text", "created_at")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.id: str = data["id"]
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster")
        self.user: PartialUser = _transform_user(client, data, "user")
        self.text: str = data["text"]
        self.created_at: datetime.datetime = _parse_datetime(data["created_at"])


class AutomodMessageHoldData(EventData):
    """
    Represents a message being held by automod for manual review.

    Attributes
    ------------
    message_id: :class:`str`
        The ID of the message.
    message_content: :class:`str`
        The contents of the message
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster from which the message was held.
    user: :class:`~twitchio.PartialUser`
        The user that sent the message.
    level: :class:`int`
        The level of alarm raised for this message.
    category: :class:`str`
        The category of alarm that was raised for this message.
    created_at: :class:`datetime.datetime`
        When this message was held.
    message_fragments: :class:`dict`
        The fragments of this message. This includes things such as emotes and cheermotes. An example from twitch is provided:

        .. code:: json

            {
                "emotes": [
                    {
                        "text": "badtextemote1",
                        "id": "emote-123",
                        "set-id": "set-emote-1"
                    },
                    {
                        "text": "badtextemote2",
                        "id": "emote-234",
                        "set-id": "set-emote-2"
                    }
                ],
                "cheermotes": [
                    {
                        "text": "badtextcheermote1",
                        "amount": 1000,
                        "prefix": "prefix",
                        "tier": 1
                    }
                ]
            }
    """

    __slots__ = (
        "message_id",
        "message_content",
        "broadcaster",
        "user",
        "level",
        "category",
        "message_fragments",
        "created_at",
    )

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.message_id: str = data["message_id"]
        self.message_content: str = data["message"]
        self.message_fragments: Dict[str, Dict[str, Any]] = data["fragments"]
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.user: PartialUser = _transform_user(client, data, "user")
        self.level: int = data["level"]
        self.category: str = data["category"]
        self.created_at: datetime.datetime = _parse_datetime(data["held_at"])


class AutomodMessageUpdateData(EventData):
    """
    Represents a message that was updated by a moderator in the automod queue.

    Attributes
    ------------
    message_id: :class:`str`
        The ID of the message.
    message_content: :class:`str`
        The contents of the message
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster from which the message was held.
    user: :class:`~twitchio.PartialUser`
        The user that sent the message.
    moderator: :class:`~twitchio.PartialUser`
        The moderator that updated the message status.
    status: :class:`str`
        The new status of the message. Typically one of ``approved`` or ``denied``.
    level: :class:`int`
        The level of alarm raised for this message.
    category: :class:`str`
        The category of alarm that was raised for this message.
    created_at: :class:`datetime.datetime`
        When this message was held.
    message_fragments: :class:`dict`
        The fragments of this message. This includes things such as emotes and cheermotes. An example from twitch is provided:

        .. code:: json

            {
                "emotes": [
                    {
                        "text": "badtextemote1",
                        "id": "emote-123",
                        "set-id": "set-emote-1"
                    },
                    {
                        "text": "badtextemote2",
                        "id": "emote-234",
                        "set-id": "set-emote-2"
                    }
                ],
                "cheermotes": [
                    {
                        "text": "badtextcheermote1",
                        "amount": 1000,
                        "prefix": "prefix",
                        "tier": 1
                    }
                ]
            }
    """

    __slots__ = (
        "message_id",
        "message_content",
        "broadcaster",
        "user",
        "moderator",
        "level",
        "category",
        "message_fragments",
        "created_at",
        "status",
    )

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.message_id: str = data["message_id"]
        self.message_content: str = data["message"]
        self.message_fragments: Dict[str, Dict[str, Any]] = data["fragments"]
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.moderator: PartialUser = _transform_user(client, data, "moderator_user")
        self.user: PartialUser = _transform_user(client, data, "user")
        self.level: int = data["level"]
        self.category: str = data["category"]
        self.created_at: datetime.datetime = _parse_datetime(data["held_at"])
        self.status: str = data["status"]


class AutomodSettingsUpdateData(EventData):
    """
    Represents a channels automod settings being updated.

    Attributes
    ------------
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster for which the settings were updated.
    moderator: :class:`~twitchio.PartialUser`
        The moderator that updated the settings.
    overall :class:`int` | ``None``
        The overall level of automod aggressiveness.
    disability: :class:`int` | ``None``
        The aggression towards disability.
    aggression: :class:`int` | ``None``
        The aggression towards aggressive users.
    sex: :class:`int` | ``None``
        The aggression towards sexuality/gender.
    misogyny: :class:`int` | ``None``
        The aggression towards misogyny.
    bullying: :class:`int` | ``None``
        The aggression towards bullying.
    swearing: :class:`int` | ``None``
        The aggression towards cursing/language.
    race_religion: :class:`int` | ``None``
        The aggression towards race, ethnicity, and religion.
    sexual_terms: :class:`int` | ``None``
        The aggression towards sexual terms/references.
    """

    __slots__ = (
        "broadcaster",
        "moderator",
        "overall",
        "disability",
        "aggression",
        "sex",
        "misogyny",
        "bullying",
        "swearing",
        "race_religion",
        "sexual_terms",
    )

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.moderator: PartialUser = _transform_user(client, data, "moderator_user")
        self.overall: Optional[int] = data["overall"]
        self.disability: Optional[int] = data["disability"]
        self.aggression: Optional[int] = data["aggression"]
        self.sex: Optional[int] = data["sex"]
        self.misogyny: Optional[int] = data["misogyny"]
        self.bullying: Optional[int] = data["bullying"]
        self.swearing: Optional[int] = data["swearing"]
        self.race_religion: Optional[int] = data["race_ethnicity_or_religion"]
        self.sexual_terms: Optional[int] = data["sex_based_terms"]


class AutomodTermsUpdateData(EventData):
    """
    Represents a channels automod terms being updated.

    .. note::

        Private terms are not sent.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The broadcaster for which the terms were updated.
    moderator: :class:`~twitchio.PartialUser`
        The moderator who updated the terms.
    action: :class:`str`
        The action type.
    from_automod: :class:`bool`
        Whether the action was taken by automod.
    terms: List[:class:`str`]
        The terms that were applied.
    """

    __slots__ = ("broadcaster", "moderator", "action", "from_automod", "terms")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.moderator: PartialUser = _transform_user(client, data, "moderator_user")
        self.action: str = data["action"]
        self.from_automod: bool = data["from_automod"]
        self.terms: List[str] = data["terms"]


class ChannelModerateData(EventData):
    """
    Represents a channel moderation event.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The channel where the moderation event occurred.
    moderator: :class:`~twitchio.PartialUser`
        The moderator who performed the action.
    action: :class:`str`
        The action performed.
    followers: Optional[:class:`Followers`]
        Metadata associated with the followers command.
    slow: Optional[:class:`Slow`]
        Metadata associated with the slow command.
    vip: Optional[:class:`VIPStatus`]
        Metadata associated with the vip command.
    unvip: Optional[:class:`VIPStatus`]
        Metadata associated with the vip command.
    mod: Optional[:class:`ModeratorStatus`]
        Metadata associated with the mod command.
    unmod: Optional[:class:`ModeratorStatus`]
        Metadata associated with the mod command.
    ban: Optional[:class:`BanStatus`]
        Metadata associated with the ban command.
    unban: Optional[:class:`BanStatus`]
        Metadata associated with the unban command.
    timeout: Optional[:class:`TimeoutStatus`]
        Metadata associated with the timeout command.
    untimeout: Optional[:class:`TimeoutStatus`]
        Metadata associated with the untimeout command.
    raid: Optional[:class:`RaidStatus`]
        Metadata associated with the raid command.
    unraid: Optional[:class:`RaidStatus`]
        Metadata associated with the unraid command.
    delete: Optional[:class:`Delete`]
        Metadata associated with the delete command.
    automod_terms: Optional[:class:`AutoModTerms`]
        Metadata associated with the automod terms changes.
    unban_request: Optional[:class:`UnBanRequest`]
        Metadata associated with an unban request.
    """

    __slots__ = (
        "broadcaster",
        "moderator",
        "action",
        "followers",
        "slow",
        "vip",
        "unvip",
        "mod",
        "unmod",
        "ban",
        "unban",
        "timeout",
        "untimeout",
        "raid",
        "unraid",
        "delete",
        "automod_terms",
        "unban_request",
    )

    class Followers:
        """
        Metadata associated with the followers command.

        Attributes
        -----------
        follow_duration_minutes: :class:`int`
            The length of time, in minutes, that the followers must have followed the broadcaster to participate in the chat room.
        """

        def __init__(self, data: dict) -> None:
            self.follow_duration_minutes: int = data["follow_duration_minutes"]

    class Slow:
        """
        Metadata associated with the slow command.

        Attributes
        -----------
        wait_time_seconds: :class:`int`
            The amount of time, in seconds, that users need to wait between sending messages.
        """

        def __init__(self, data: dict) -> None:
            self.wait_time_seconds: int = data["wait_time_seconds"]

    class VIPStatus:
        """
        Metadata associated with the vip / unvip command.

        Attributes
        -----------
        user: :class:`~twitchio.PartialUser`
            The user who is gaining or losing VIP access.
        """

        def __init__(self, client: EventSubClient, data: dict) -> None:
            self.user: PartialUser = _transform_user(client, data, "user")

    class ModeratorStatus:
        """
        Metadata associated with the mod / unmod command.

        Attributes
        -----------
        user: :class:`~twitchio.PartialUser`
            The user who is gaining or losing moderator access.
        """

        def __init__(self, client: EventSubClient, data: dict) -> None:
            self.user: PartialUser = _transform_user(client, data, "user")

    class BanStatus:
        """
        Metadata associated with the ban / unban command.

        Attributes
        -----------
        user: :class:`~twitchio.PartialUser`
            The user who is banned / unbanned.
        reason: Optional[:class:`str`]
            Reason for the ban.
        """

        def __init__(self, client: EventSubClient, data: dict) -> None:
            self.user: PartialUser = _transform_user(client, data, "user")
            self.reason: Optional[str] = data.get("reason")

    class TimeoutStatus:
        """
        Metadata associated with the timeout / untimeout command.

        Attributes
        -----------
        user: :class:`~twitchio.PartialUser`
            The user who is timedout / untimedout.
        reason: Optional[:class:`str`]
            Reason for the timeout.
        expires_at: Optional[:class:`datetime.datetime`]
            Datetime the timeout expires.
        """

        def __init__(self, client: EventSubClient, data: dict) -> None:
            self.user: PartialUser = _transform_user(client, data, "user")
            self.reason: Optional[str] = data.get("reason")
            self.expires_at: Optional[datetime.datetime] = (
                _parse_datetime(data["expires_at"]) if data.get("expires_at") is not None else None
            )

    class RaidStatus:
        """
        Metadata associated with the raid / unraid command.

        Attributes
        -----------
        user: :class:`~twitchio.PartialUser`
            The user who is timedout / untimedout.
        viewer_count: :class:`int`
            The viewer count.
        """

        def __init__(self, client: EventSubClient, data: dict) -> None:
            self.user: PartialUser = _transform_user(client, data, "user")
            self.viewer_count: int = data["viewer_count"]

    class Delete:
        """
        Metadata associated with the delete command.

        Attributes
        -----------
        user: :class:`~twitchio.PartialUser`
            The user who is timedout / untimedout.
        message_id: :class:`str`
            The id of deleted message.
        message_body: :class:`str`
            The message body of the deleted message.
        """

        def __init__(self, client: EventSubClient, data: dict) -> None:
            self.user: PartialUser = _transform_user(client, data, "user")
            self.message_id: str = data["message_id"]
            self.message_body: str = data["message_body"]

    class AutoModTerms:
        """
        Metadata associated with the automod terms change.

        Attributes
        -----------
        action: :class:`Literal["add", "remove"]`
            Either “add” or “remove”.
        list: :class:`Literal["blocked", "permitted"]`
            Either “blocked” or “permitted”.
        terms: List[:class:`str`]
            Terms being added or removed.
        from_automod: :class:`bool`
            Whether the terms were added due to an Automod message approve/deny action.
        """

        def __init__(self, data: dict) -> None:
            self.action: Literal["add", "remove"] = data["action"]
            self.list: Literal["blocked", "permitted"] = data["list"]
            self.terms: List[str] = data["terms"]
            self.from_automod: bool = data["from_automod"]

    class UnBanRequest:
        """
        Metadata associated with the slow command.

        Attributes
        -----------
        user: :class:`~twitchio.PartialUser`
            The user who is requesting an unban.
        is_approved: :class:`bool`
            Whether or not the unban request was approved or denied.
        moderator_message: :class:`str`
            The message included by the moderator explaining their approval or denial.
        """

        def __init__(self, client: EventSubClient, data: dict) -> None:
            self.user: PartialUser = _transform_user(client, data, "user")
            self.is_approved: bool = data["is_approved"]
            self.moderator_message: str = data["moderator_message"]

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster = _transform_user(client, data, "broadcaster_user")
        self.moderator = _transform_user(client, data, "moderator_user")
        self.action: str = data["action"]
        self.followers = self.Followers(data["followers"]) if data.get("followers") is not None else None
        self.slow = self.Slow(data["slow"]) if data.get("slow") is not None else None
        self.vip = self.VIPStatus(client, data["vip"]) if data.get("vip") is not None else None
        self.unvip = self.VIPStatus(client, data["unvip"]) if data.get("unvip") is not None else None
        self.mod = self.ModeratorStatus(client, data["mod"]) if data.get("mod") is not None else None
        self.unmod = self.ModeratorStatus(client, data["unmod"]) if data.get("unmod") is not None else None
        self.ban = self.BanStatus(client, data["ban"]) if data.get("ban") is not None else None
        self.unban = self.BanStatus(client, data["unban"]) if data.get("unban") is not None else None
        self.timeout = self.TimeoutStatus(client, data["timeout"]) if data.get("timeout") is not None else None
        self.untimeout = self.TimeoutStatus(client, data["untimeout"]) if data.get("untimeout") is not None else None
        self.raid = self.RaidStatus(client, data["raid"]) if data.get("raid") is not None else None
        self.unraid = self.RaidStatus(client, data["unraid"]) if data.get("unraid") is not None else None
        self.delete = self.Delete(client, data["delete"]) if data.get("delete") is not None else None
        self.automod_terms = self.AutoModTerms(data["automod_terms"]) if data.get("automod_terms") is not None else None
        self.unban_request = (
            self.UnBanRequest(client, data["unban_request"]) if data.get("unban_request") is not None else None
        )


class SuspiciousUserUpdateData(EventData):
    """
    Represents a suspicious user update event.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The channel where the treatment for a suspicious user was updated.
    moderator: :class:`~twitchio.PartialUser`
        The moderator who updated the terms.
    user: :class:`~twitchio.PartialUser`
        The the user that sent the message.
    trust_status: :class:`Literal["active_monitoring", "restricted", "none"]`
        The status set for the suspicious user. Can be the following: “none”, “active_monitoring”, or “restricted”.
    """

    __slots__ = ("broadcaster", "moderator", "user", "trust_status")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster_user")
        self.moderator: PartialUser = _transform_user(client, data, "moderator_user")
        self.user: PartialUser = _transform_user(client, data, "user")
        self.trust_status: Literal["active_monitoring", "restricted", "none"] = data["low_trust_status"]


class AutoCustomReward:
    """
    A reward object for an Auto Reward Redeem.

    Attributes
    -----------
    type: :class:`str`
        The type of the reward. One of ``single_message_bypass_sub_mode``, ``send_highlighted_message``, ``random_sub_emote_unlock``,
        ``chosen_sub_emote_unlock``, ``chosen_modified_sub_emote_unlock``, ``message_effect``, ``gigantify_an_emote``, ``celebration``.
    cost: :class:`int`
        How much the reward costs.
    unlocked_emote_id: Optional[:class:`str`]
        The unlocked emote, if applicable.
    unlocked_emote_name: Optional[:class:`str`]
        The unlocked emote, if applicable.
    """

    def __init__(self, data: dict):
        self.type: str = data["type"]
        self.cost: int = data["cost"]
        self.unlocked_emote_id: Optional[str] = data["unlocked_emote"] and data["unlocked_emote"]["id"]
        self.unlocked_emote_name: Optional[str] = data["unlocked_emote"] and data["unlocked_emote"]["name"]


class AutoRewardRedeem(EventData):
    """
    Represents an automatic reward redemption.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The channel where the reward was redeemed.
    user: :class:`~twitchio.PartialUser`
        The user that redeemed the reward.
    id: :class:`str`
        The ID of the redemption.
    reward: :class:`AutoCustomReward`
        The reward that was redeemed.
    message: :class:`str`
        The message the user sent.
    message_emotes: :class:`dict`
        The emote data for the message.
    user_input: Optional[:class:`str`]
        The input to the reward, if it requires any.
    redeemed_at: :class:`datetime.datetime`
        When the reward was redeemed.
    """

    __slots__ = ("broadcaster", "user", "id", "reward", "message", "message_emotes", "user_input", "redeemed_at")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster")
        self.user: PartialUser = _transform_user(client, data, "user")
        self.id: str = data["id"]
        self.reward = AutoCustomReward(data["reward"])
        self.message: str = data["message"]
        self.message_emotes: dict = data["message_emotes"]
        self.user_input: Optional[str] = data["user_input"]
        self.redeemed_at: datetime.datetime = _parse_datetime(data["redeemed_at"])


class ChannelVIPAddRemove(EventData):
    """
    Represents a VIP being added/removed from a channel.

    Attributes
    -----------
    broadcaster: :class:`~twitchio.PartialUser`
        The channel that the VIP was added/removed from.
    user: :class:`~twitchio.PartialUser`
        The user that was added/removed as a VIP.
    """

    __slots__ = ("broadcaster", "user")

    def __init__(self, client: EventSubClient, data: dict) -> None:
        self.broadcaster: PartialUser = _transform_user(client, data, "broadcaster")
        self.user: PartialUser = _transform_user(client, data, "user")


_DataType = Union[
    ChannelBanData,
    ChannelUnbanData,
    ChannelSubscribeData,
    ChannelSubscriptionEndData,
    ChannelSubscriptionGiftData,
    ChannelSubscriptionMessageData,
    ChannelCheerData,
    ChannelUpdateData,
    ChannelFollowData,
    ChannelRaidData,
    ChannelModeratorAddRemoveData,
    ChannelGoalBeginProgressData,
    ChannelGoalEndData,
    CustomRewardAddUpdateRemoveData,
    CustomRewardRedemptionAddUpdateData,
    HypeTrainBeginProgressData,
    HypeTrainEndData,
    PollBeginProgressData,
    PollEndData,
    PredictionBeginProgressData,
    PredictionLockData,
    PredictionEndData,
    StreamOnlineData,
    StreamOfflineData,
    UserAuthorizationGrantedData,
    UserAuthorizationRevokedData,
    UserUpdateData,
    ChannelShieldModeBeginData,
    ChannelShieldModeEndData,
    ChannelShoutoutCreateData,
    ChannelShoutoutReceiveData,
    ChannelCharityDonationData,
    ChannelUnbanRequestCreateData,
    ChannelUnbanRequestResolveData,
    AutomodMessageHoldData,
    AutomodMessageUpdateData,
    AutomodSettingsUpdateData,
    AutomodTermsUpdateData,
    SuspiciousUserUpdateData,
    ChannelModerateData,
    AutoRewardRedeem,
    ChannelVIPAddRemove,
]


class _SubTypesMeta(type):
    def __new__(mcs, clsname, bases, attributes):
        attributes["_type_map"] = {args[0]: args[2] for name, args in attributes.items() if not name.startswith("_")}
        attributes["_name_map"] = {args[0]: name for name, args in attributes.items() if not name.startswith("_")}
        return super().__new__(mcs, clsname, bases, attributes)


class _SubscriptionTypes(metaclass=_SubTypesMeta):
    _type_map: Dict[str, Type[_DataType]]
    _name_map: Dict[str, str]

    automod_message_hold = "automod.message.hold", 1, AutomodMessageHoldData
    automod_message_update = "automod.message.update", 1, AutomodMessageUpdateData
    automod_settings_update = "automod.settings.update", 1, AutomodSettingsUpdateData
    automod_terms_update = "automod.terms.update", 1, AutomodTermsUpdateData

    follow = "channel.follow", 1, ChannelFollowData
    followV2 = "channel.follow", 2, ChannelFollowData
    subscription = "channel.subscribe", 1, ChannelSubscribeData
    subscription_end = "channel.subscription.end", 1, ChannelSubscriptionEndData
    subscription_gift = "channel.subscription.gift", 1, ChannelSubscriptionGiftData
    subscription_message = "channel.subscription.message", 1, ChannelSubscriptionMessageData
    cheer = "channel.cheer", 1, ChannelCheerData
    raid = "channel.raid", 1, ChannelRaidData
    ban = "channel.ban", 1, ChannelBanData
    unban = "channel.unban", 1, ChannelUnbanData

    channel_update = "channel.update", 1, ChannelUpdateData
    channel_moderator_add = "channel.moderator.add", 1, ChannelModeratorAddRemoveData
    channel_moderator_remove = "channel.moderator.remove", 1, ChannelModeratorAddRemoveData
    channel_reward_add = "channel.channel_points_custom_reward.add", 1, CustomRewardAddUpdateRemoveData
    channel_reward_update = "channel.channel_points_custom_reward.update", 1, CustomRewardAddUpdateRemoveData
    channel_reward_remove = "channel.channel_points_custom_reward.remove", 1, CustomRewardAddUpdateRemoveData
    channel_reward_redeem = (
        "channel.channel_points_custom_reward_redemption.add",
        1,
        CustomRewardRedemptionAddUpdateData,
    )
    channel_reward_redeem_updated = (
        "channel.channel_points_custom_reward_redemption.update",
        1,
        CustomRewardRedemptionAddUpdateData,
    )

    auto_reward_redeem = "channel.channel_points_automatic_reward_redemption.add", 1, AutoRewardRedeem

    channel_goal_begin = "channel.goal.begin", 1, ChannelGoalBeginProgressData
    channel_goal_progress = "channel.goal.progress", 1, ChannelGoalBeginProgressData
    channel_goal_end = "channel.goal.end", 1, ChannelGoalEndData

    channel_shield_mode_begin = "channel.shield_mode.begin", 1, ChannelShieldModeBeginData
    channel_shield_mode_end = "channel.shield_mode.end", 1, ChannelShieldModeEndData

    channel_shoutout_create = "channel.shoutout.create", 1, ChannelShoutoutCreateData
    channel_shoutout_receive = "channel.shoutout.receive", 1, ChannelShoutoutReceiveData

    channel_charity_donate = "channel.charity_campaign.donate", 1, ChannelCharityDonationData

    channel_moderate = "channel.moderate", 1, ChannelModerateData

    hypetrain_begin = "channel.hype_train.begin", 1, HypeTrainBeginProgressData
    hypetrain_progress = "channel.hype_train.progress", 1, HypeTrainBeginProgressData
    hypetrain_end = "channel.hype_train.end", 1, HypeTrainEndData

    poll_begin = "channel.poll.begin", 1, PollBeginProgressData
    poll_progress = "channel.poll.progress", 1, PollBeginProgressData
    poll_end = "channel.poll.end", 1, PollEndData

    prediction_begin = "channel.prediction.begin", 1, PredictionBeginProgressData
    prediction_progress = "channel.prediction.progress", 1, PredictionBeginProgressData
    prediction_lock = "channel.prediction.lock", 1, PredictionLockData
    prediction_end = "channel.prediction.end", 1, PredictionEndData

    stream_start = "stream.online", 1, StreamOnlineData
    stream_end = "stream.offline", 1, StreamOfflineData

    unban_request_create = "channel.unban_request.create", 1, ChannelUnbanRequestCreateData
    unban_request_resolve = "channel.unban_request.resolve", 1, ChannelUnbanRequestResolveData

    channel_vip_add = "channel.vip.add", 1, ChannelVIPAddRemove
    channel_vip_remove = "channel.vip.remove", 1, ChannelVIPAddRemove

    user_authorization_grant = "user.authorization.grant", 1, UserAuthorizationGrantedData
    user_authorization_revoke = "user.authorization.revoke", 1, UserAuthorizationRevokedData

    user_update = "user.update", 1, UserUpdateData

    suspicious_user_update = "channel.suspicious_user.update", 1, SuspiciousUserUpdateData


SubscriptionTypes = _SubscriptionTypes()


class TransportType(Enum):
    webhook = "webhook"
    websocket = "websocket"



"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import time
import functools
from typing import Hashable


__all__ = ("TimedCache", "user_cache", "id_cache")


class TimedCache(dict):
    def __init__(self, seconds: int):
        self.__timeout = seconds
        super().__init__()

    def _verify_cache(self):
        now = time.monotonic()
        to_remove = [key for (key, (_, exp)) in self.items() if now > (exp + self.__timeout)]
        for k in to_remove:
            del self[k]

    def __getitem__(self, key):
        self._verify_cache()
        return super().__getitem__(key)[0]

    def __setitem__(self, key, value):
        super().__setitem__(key, (value, time.monotonic()))

    def __contains__(self, key):
        self._verify_cache()
        return {a: b[0] for a, b in self.items()}.__contains__(key)


def user_cache(timer=300):
    cache = TimedCache(timer)

    def wraps(func):
        @functools.wraps(func)
        async def _wraps(cls, names: list = None, ids: list = None, force=False, token=None):
            if not force:
                existing = []
                if names:
                    existing.extend([cache[x] for x in names if x in cache])

                if ids:
                    existing.extend([cache[x] for x in ids if x in cache])

                if len(existing) == (len(names) if names else 0) + (len(ids) if ids else 0):
                    return existing

            values = await func(cls, names=names, ids=ids, token=token)
            for v in values:
                cache[v.id] = v
                cache[v.name] = v

            return values

        return _wraps

    return wraps


def id_cache(timer=300):
    cache = TimedCache(timer)

    def wraps(func):
        @functools.wraps(func)
        def _wraps(cls, id: Hashable):
            if id in cache:
                return cache[id]

            value = func(cls, id)
            if value is not None:
                cache[id] = value

            return value

        return _wraps

    return wraps



"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
import inspect

import itertools
import copy
import types
from typing import Any, Union, Optional, Callable, Awaitable, Tuple, TYPE_CHECKING, List, Type, Set, TypeVar
from typing_extensions import Literal

from twitchio.abcs import Messageable
from .cooldowns import *
from .errors import *
from . import builtin_converter

if TYPE_CHECKING:
    import sys

    from twitchio import Message, Chatter, PartialChatter, Channel, User, PartialUser
    from . import Cog, Bot
    from .stringparser import StringParser

    if sys.version_info >= (3, 10):
        UnionT = Union[types.UnionType, Union]
    else:
        UnionT = Union


__all__ = ("Command", "command", "Group", "Context", "cooldown")


class EmptyArgumentSentinel:
    def __repr__(self) -> str:
        return "<EMPTY>"

    def __eq__(self, __value: object) -> bool:
        return False


EMPTY = EmptyArgumentSentinel()


def _boolconverter(_, param: str):
    param = param.lower()
    if param in {"yes", "y", "1", "true", "on"}:
        return True
    elif param in {"no", "n", "0", "false", "off"}:
        return False
    raise BadArgument(f"Expected a boolean value, got {param}")


class Command:
    """A class for implementing bot commands.

    Parameters
    ------------
    name: :class:`str`
        The name of the command.
    func: :class:`Callable`
        The coroutine that executes when the command is invoked.

    Attributes
    ------------
    name: :class:`str`
        The name of the command.
    cog: :class:`~twitchio.ext.commands.Cog`
        The cog this command belongs to.
    aliases: Optional[Union[:class:`list`, :class:`tuple`]]
        Aliases that can be used to also invoke the command.
    """

    def __init__(self, name: str, func: Callable, **attrs) -> None:
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Command callback must be a coroutine.")
        self._callback = func
        self._checks = []
        self._cooldowns = []
        self._name = name

        self._instance = None
        self.cog = None
        self.parent: Optional[Group] = attrs.get("parent")

        try:
            self._checks.extend(func.__checks__)  # type: ignore
        except AttributeError:
            pass
        try:
            self._cooldowns.extend(func.__cooldowns__)  # type: ignore
        except AttributeError:
            pass
        self.aliases = attrs.get("aliases", None)
        sig = inspect.signature(func)
        self.params = sig.parameters.copy()  # type: ignore

        self.event_error = None
        self._before_invoke = None
        self._after_invoke = None
        self.no_global_checks = attrs.get("no_global_checks", False)

        for key, value in self.params.items():
            if isinstance(value.annotation, str):
                self.params[key] = value.replace(annotation=eval(value.annotation, func.__globals__))  # type: ignore

    @property
    def name(self) -> str:
        return self._name

    @property
    def full_name(self) -> str:
        if not self.parent:
            return self._name
        return f"{self.parent.full_name} {self._name}"

    def _is_optional_argument(self, converter: Any):
        return (getattr(converter, "__origin__", None) is Union or isinstance(converter, types.UnionType)) and type(
            None
        ) in converter.__args__

    def resolve_union_callback(self, name: str, converter: UnionT) -> Callable[[Context, str], Any]:
        # print(type(converter), converter.__args__)

        args = converter.__args__  # type: ignore # pyright doesnt like this

        async def _resolve(context: Context, arg: str) -> Any:
            t = EMPTY

            for original in args:
                underlying = self._resolve_converter(name, original, context)

                try:
                    t: Any = underlying(context, arg)
                    if inspect.iscoroutine(t):
                        t = await t

                    break
                except Exception as l:
                    t = EMPTY  # thisll get changed when t is a coroutine, but is still invalid, so roll it back
                    continue

            if t is EMPTY:
                raise UnionArgumentParsingFailed(name, args)

            return t

        return _resolve

    def resolve_optional_callback(self, name: str, converter: Any, context: Context) -> Callable[[Context, str], Any]:
        underlying = self._resolve_converter(name, converter.__args__[0], context)

        async def _resolve(context: Context, arg: str) -> Any:
            try:
                t: Any = underlying(context, arg)
                if inspect.iscoroutine(t):
                    t = await t

            except Exception:
                return EMPTY  # instruct the parser to roll back and ignore this argument

            return t

        return _resolve

    def _resolve_converter(
        self, name: str, converter: Union[Callable, Awaitable, type], ctx: Context
    ) -> Callable[..., Any]:
        if (
            isinstance(converter, type)
            and converter.__module__.startswith("twitchio")
            and converter in builtin_converter._mapping
        ):
            return self._convert_builtin_type(name, converter, builtin_converter._mapping[converter])

        elif converter is bool:
            converter = self._convert_builtin_type(name, bool, _boolconverter)

        elif converter in (str, int):
            original: type[str | int] = converter  # type: ignore
            converter = self._convert_builtin_type(name, original, lambda _, arg: original(arg))

        elif self._is_optional_argument(converter):
            return self.resolve_optional_callback(name, converter, ctx)

        elif isinstance(converter, types.UnionType) or getattr(converter, "__origin__", None) is Union:
            return self.resolve_union_callback(name, converter)  # type: ignore

        elif hasattr(converter, "__metadata__"):  # Annotated
            annotated = converter.__metadata__  # type: ignore
            return self._resolve_converter(name, annotated[0], ctx)

        return converter  # type: ignore

    def _convert_builtin_type(
        self,
        arg_name: str,
        original: type,
        converter: Union[Callable[[Context, str], Any], Callable[[Context, str], Awaitable[Any]]],
    ) -> Callable[[Context, str], Awaitable[Any]]:
        async def resolve(ctx, arg: str) -> Any:
            try:
                t = converter(ctx, arg)

                if inspect.iscoroutine(t):
                    t = await t

                return t
            except Exception as e:
                raise ArgumentParsingFailed(
                    f"Failed to convert `{arg}` to expected type {original.__name__} for argument `{arg_name}`",
                    original=e,
                    argname=arg_name,
                    expected=original,
                ) from e

        return resolve

    async def _convert_types(self, context: Context, param: inspect.Parameter, parsed: str) -> Any:
        converter = param.annotation

        if converter is param.empty:
            if param.default in (param.empty, None):
                converter = str
            else:
                converter = type(param.default)

        true_converter = self._resolve_converter(param.name, converter, context)

        try:
            argument = true_converter(context, parsed)
            if inspect.iscoroutine(argument):
                argument = await argument
        except BadArgument as e:
            if e.name is None:
                e.name = param.name

            raise
        except Exception as e:
            raise ArgumentParsingFailed(
                f"Failed to parse `{parsed}` for argument {param.name}", original=e, argname=param.name, expected=None
            ) from e
        return argument

    async def parse_args(self, context: Context, instance: Optional[Cog], parsed: dict, index=0) -> Tuple[list, dict]:
        if isinstance(self, Group):
            parsed = parsed.copy()
        iterator = iter(self.params.items())
        args = []
        kwargs = {}

        try:
            next(iterator)
            if instance:
                next(iterator)
        except StopIteration:
            raise TwitchCommandError("self or ctx is a required argument which is missing.")
        for _, param in iterator:
            index += 1
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                try:
                    argument = parsed.pop(index)
                except (KeyError, IndexError):
                    if self._is_optional_argument(param.annotation):  # parameter is optional and at the end.
                        args.append(param.default if param.default is not param.empty else None)
                        continue

                    if param.default is param.empty:
                        raise MissingRequiredArgument(argname=param.name)

                    args.append(param.default)
                else:
                    _parsed_arg = await self._convert_types(context, param, argument)

                    if _parsed_arg is EMPTY:
                        parsed[index] = argument
                        index -= 1
                        args.append(param.default if param.default is not param.empty else None)

                        continue
                    else:
                        args.append(_parsed_arg)

            elif param.kind == param.KEYWORD_ONLY:
                rest = " ".join(parsed.values())
                if rest.startswith(" "):
                    rest = rest.lstrip(" ")
                if rest:
                    rest = await self._convert_types(context, param, rest)
                elif param.default is param.empty:
                    raise MissingRequiredArgument(argname=param.name)
                else:
                    rest = param.default
                kwargs[param.name] = rest
                parsed.clear()
                break
            elif param.kind == param.VAR_POSITIONAL:
                args.extend([await self._convert_types(context, param, argument) for argument in parsed.values()])
                parsed.clear()
                break
        if parsed:
            pass  # TODO Raise Too Many Arguments.
        return args, kwargs

    async def invoke(self, context: Context, *, index=0) -> None:
        # TODO Docs
        if not context.view:
            return

        async def try_run(func, *, to_command=False):
            try:
                await func
            except Exception as _e:
                if not to_command:
                    context.bot.run_event("error", _e)
                else:
                    context.bot.run_event("command_error", context, _e)

        try:
            args, kwargs = await self.parse_args(context, self._instance, context.view.words, index=index)
        except (MissingRequiredArgument, BadArgument) as e:
            if self.event_error:
                args_ = [self._instance, context] if self._instance else [context]
                await try_run(self.event_error(*args_, e))

            context.bot.run_event("command_error", context, e)
            return

        context.args, context.kwargs = args, kwargs
        check_result = await self.handle_checks(context)

        if check_result is not True:
            context.bot.run_event("command_error", context, check_result)
            return
        limited = self._run_cooldowns(context)

        if limited:
            context.bot.run_event("command_error", context, limited[0])
            return
        instance = self._instance
        args = [instance, context] if instance else [context]
        await try_run(context.bot.global_before_invoke(context))

        if self._before_invoke:
            await try_run(self._before_invoke(*args), to_command=True)
        try:
            await self._callback(*args, *context.args, **context.kwargs)
        except Exception as e:
            if self.event_error:
                await try_run(self.event_error(*args, e))
            context.bot.run_event("command_error", context, e)
        else:
            context.bot.run_event("command_complete", context)
        # Invoke our after command hooks
        if self._after_invoke:
            await try_run(self._after_invoke(*args), to_command=True)
        await try_run(context.bot.global_after_invoke(context))

    def _run_cooldowns(self, context: Context) -> Optional[List[CommandOnCooldown]]:
        try:
            buckets = self._cooldowns[0].get_buckets(context)
        except IndexError:
            return None
        expired = []

        try:
            for bucket in buckets:
                bucket.update_bucket(context)
        except CommandOnCooldown as e:
            expired.append(e)
        return expired

    async def handle_checks(self, context: Context) -> Union[Literal[True], Exception]:
        # TODO Docs

        if not self.no_global_checks:
            checks = [predicate for predicate in itertools.chain(context.bot._checks, self._checks)]
        else:
            checks = self._checks
        try:
            for predicate in checks:
                result = predicate(context)

                if inspect.isawaitable(result):
                    result = await result  # type: ignore
                if not result:
                    raise CheckFailure(f"The check {predicate} for command {self.name} failed.")
            if self.cog and not await self.cog.cog_check(context):
                raise CheckFailure(f"The cog check for command <{self.name}> failed.")
            return True
        except Exception as e:
            return e

    async def __call__(self, context: Context, *, index=0) -> None:
        await self.invoke(context, index=index)


class Group(Command):
    def __init__(self, *args, invoke_with_subcommand=False, **kwargs) -> None:
        super(Group, self).__init__(*args, **kwargs)
        self._sub_commands = {}
        self._invoke_with_subcommand = invoke_with_subcommand

    async def __call__(self, context: Context, *, index=0) -> None:
        if not context.view:
            return
        if not context.view.words:
            return await self.invoke(context, index=index)
        arg: Tuple[int, str] = list(context.view.words.items())[0]  # type: ignore
        if arg[1] in self._sub_commands:
            _ctx = copy.copy(context)
            _ctx.view = _ctx.view.copy()
            _ctx.view.words.pop(arg[0])
            await self._sub_commands[arg[1]](_ctx, index=arg[0])

            if self._invoke_with_subcommand:
                await self.invoke(context, index=index)
        else:
            await self.invoke(context, index=index)

    def command(
        self, *, name: str = None, aliases: Union[list, tuple] = None, cls=Command, no_global_checks=False
    ) -> Callable[[Callable], Command]:
        if cls and not inspect.isclass(cls):
            raise TypeError(f"cls must be of type <class> not <{type(cls)}>")

        def decorator(func: Callable):
            fname = name or func.__name__
            cmd = cls(name=fname, func=func, aliases=aliases, no_global_checks=no_global_checks, parent=self)
            self._sub_commands[cmd.name] = cmd
            if cmd.aliases:
                for a in cmd.aliases:
                    self._sub_commands[a] = cmd
            return cmd

        return decorator

    def group(
        self,
        *,
        name: str = None,
        aliases: Union[list, tuple] = None,
        cls: Type[Group] = None,
        no_global_checks=False,
        invoke_with_subcommand=False,
    ) -> Callable[[Callable], Group]:
        cls = cls or Group
        if cls and not inspect.isclass(cls):
            raise TypeError(f"cls must be of type <class> not <{type(cls)}>")

        def decorator(func: Callable):
            fname = name or func.__name__
            cmd = cls(
                name=fname,
                func=func,
                aliases=aliases,
                no_global_checks=no_global_checks,
                parent=self,
                invoke_with_subcommand=invoke_with_subcommand,
            )
            self._sub_commands[cmd.name] = cmd
            if cmd.aliases:
                for a in cmd.aliases:
                    self._sub_commands[a] = cmd
            return cmd

        return decorator


class Context(Messageable):
    """
    A class that represents the context in which a command is being invoked under.

    This class contains the meta data to help you understand more about the invocation context.
    This class is not created manually and is instead passed around to commands as the first parameter.

    Attributes
    -----------
    message: :class:`~twitchio.Message`
        The message that triggered the command being executed.
    channel: :class:`~twitchio.Channel`
        The channel the command was invoked in.
    author: Union[:class:`~twitchio.PartialChatter`, :class:`~twitchio.Chatter`]
        The Chatter object of the user in chat that invoked the command.
    prefix: Optional[:class:`str`]
        The prefix that was used to invoke the command.
    command: Optional[:class:`~twitchio.ext.commands.Command`]
        The command that was invoked
    cog: Optional[:class:`~twitchio.ext.commands.Cog`]
        The cog that contains the command that was invoked.
    args: Optional[List[:class:`Any`]]
        List of arguments that were passed to the command.
    kwargs: Optional[Dict[:class:`str`, :class:`Any`]]
        List of kwargs that were passed to the command.
    view: Optional[:class:`~twitchio.ext.commmands.StringParser`]
        StringParser object that breaks down the command string received.
    bot: :class:`~twitchio.ext.commands.Bot`
        The bot that contains the command that was invoked.
    """

    __messageable_channel__ = True

    def __init__(self, message: Message, bot: Bot, **attrs) -> None:
        self.message: Message = message
        self.channel: Channel = message.channel
        self.author: Union[Chatter, PartialChatter] = message.author

        self.prefix: Optional[str] = attrs.get("prefix")

        self.command: Optional[Command] = attrs.get("command")
        if self.command:
            self.cog: Optional[Cog] = self.command.cog
        self.args: Optional[list] = attrs.get("args")
        self.kwargs: Optional[dict] = attrs.get("kwargs")

        self.view: Optional[StringParser] = attrs.get("view")
        self.is_valid: bool = attrs.get("valid")

        self.bot: Bot = bot
        self._ws = self.author._ws

    def _fetch_channel(self) -> Messageable:
        return self.channel or self.author  # Abstract method

    def _fetch_websocket(self):
        return self._ws  # Abstract method

    def _fetch_message(self):
        return self.message  # Abstract method

    def _bot_is_mod(self) -> bool:
        if not self.channel:
            return False
        cache = self._ws._cache[self.channel._name]
        for user in cache:
            if user.name == self._ws.nick:
                try:
                    mod = user.is_mod
                except AttributeError:
                    return False
                return mod

    @property
    def chatters(self) -> Optional[Set[Chatter]]:
        """The channels current chatters."""
        try:
            users = self._ws._cache[self.channel._name]
        except (KeyError, AttributeError):
            return None
        return users

    @property
    def users(self) -> Optional[Set[Chatter]]:  # Alias to chatters
        """Alias to chatters."""
        return self.chatters

    def get_user(self, name: str) -> Optional[Union[PartialChatter, Chatter]]:
        """Retrieve a user from the channels user cache.

        Parameters
        -----------
        name: str
            The user's name to try and retrieve.

        Returns
        --------
        Union[:class:`twitchio.Chatter`, :class:`twitchio.PartialChatter`]
            Could be a :class:`twitchio.PartialChatter` depending on how the user joined the channel.
            Returns None if no user was found.
        """
        name = name.lower()

        if not self.channel:
            return None
        cache = self._ws._cache[self.channel._name]
        for user in cache:
            if user.name == name:
                return user
        return None

    async def reply(self, content: str):
        """|coro|


        Send a message in reply to the user who sent a message in the destination
        associated with the dataclass.

        Destination will be the context of which the message/command was sent.

        Parameters
        ------------
        content: str
            The content you wish to send as a message. The content must be a string.

        Raises
        --------
        InvalidContent
            Invalid content.
        """
        entity = self._fetch_channel()
        ws = self._fetch_websocket()
        message = self._fetch_message()

        self.check_content(content)
        self.check_bucket(channel=entity.name)

        try:
            name = entity.channel.name
        except AttributeError:
            name = entity.name
        if entity.__messageable_channel__:
            await ws.reply(message.id, f"PRIVMSG #{name} :{content}\r\n")
        else:
            await ws.send(f"PRIVMSG #jtv :/w {name} {content}\r\n")


C = TypeVar("C", bound="Command")
G = TypeVar("G", bound="Group")


def command(
    *, name: str = None, aliases: Union[list, tuple] = None, cls: type[C] = Command, no_global_checks=False
) -> Callable[[Callable], C]:
    if cls and not inspect.isclass(cls):
        raise TypeError(f"cls must be of type <class> not <{type(cls)}>")

    def decorator(func: Callable) -> C:
        fname = name or func.__name__
        return cls(
            name=fname,
            func=func,
            aliases=aliases,
            no_global_checks=no_global_checks,
        )

    return decorator


def group(
    *,
    name: str = None,
    aliases: Union[list, tuple] = None,
    cls: G = Group,
    no_global_checks=False,
    invoke_with_subcommand=False,
) -> Callable[[Callable], G]:
    if cls and not inspect.isclass(cls):
        raise TypeError(f"cls must be of type <class> not <{type(cls)}>")

    def decorator(func: Callable) -> G:
        fname = name or func.__name__
        return cls(
            name=fname,
            func=func,
            aliases=aliases,
            no_global_checks=no_global_checks,
            invoke_with_subcommand=invoke_with_subcommand,
        )

    return decorator


FN = TypeVar("FN")


def cooldown(rate, per, bucket=Bucket.default):
    def decorator(func: FN) -> FN:
        if isinstance(func, Command):
            func._cooldowns.append(Cooldown(rate, per, bucket))
        else:
            func.__cooldowns__ = [Cooldown(rate, per, bucket)]
        return func

    return decorator
